import logging

from pg_discuss.app import app_factory

app = app_factory()

if __name__ == '__main__':
    app.script_manager.run()
else:
    # If not running a management command, set log level and log configuration.
    # Set up logging

    if not app.debug:
        app.logger.addHandler(logging.StreamHandler())
        app.logger.setLevel(app.config['LOGLEVEL'])

    # Log configuration parameters
    app.logger.info('Configuration parameters:\n{}'.format(
        '\n'.join([k + '=' + str(v) for k, v in
                   sorted(app.config.items())
                   if k not in app.config['DO_NOT_LOG_VARS']])))

    app.logger.info('Found extensions/drivers:\n{}'.format(
        '\n'.join([e.name for e in app.ext_mgr_all.extensions])))

    app.logger.info('Enabled IdentityPolicy driver:\n{}'.format(
        app.identity_policy_loader.extensions[0].name))

    app.logger.info('Enabled CommentRenderer driver:\n{}'.format(
        app.comment_renderer_loader.extensions[0].name))

    app.logger.info('Enabled JSONEncoder driver:\n{}'.format(
        app.json_encoder_loader.extensions[0].name))


    app.logger.info('Enabled extensions:\n{}'.format(
        '\n'.join([e.name for e in app.ext_mgr.extensions])))
    # Make `application` alias for uwsgi/pypy
    # See https://github.com/unbit/uwsgi/issues/900
    application = app



"""Form for Admin login and a Flask-Script command to set up an Admin user from
the command line.
"""
import getpass

import flask_script
import flask_wtf
import werkzeug
import wtforms

from . import models
from .db import db


class LoginForm(flask_wtf.Form):
    """Login form for Admin users."""
    login = wtforms.fields.TextField(
        validators=[wtforms.validators.DataRequired()])
    password = wtforms.fields.PasswordField(
        validators=[wtforms.validators.DataRequired()])

    def validate_login(self, field):
        """Validate that user is valid and password matches stored hash."""
        user = self.get_user()

        if user is None:
            raise wtforms.validators.ValidationError('Invalid user')

        if not werkzeug.check_password_hash(user.password, self.password.data):
            raise wtforms.validators.ValidationError('Invalid password')

    def get_user(self):
        """Get the user object associated with the login."""
        return (
            db.session.query(models.AdminUser)
            .filter_by(login=self.login.data).first()
        )


class RegistrationForm(wtforms.Form):
    """Registration form to validate new Admin user data."""
    login = wtforms.fields.TextField(
        validators=[wtforms.validators.DataRequired()])
    email = wtforms.fields.TextField()
    password = wtforms.fields.PasswordField(
        validators=[wtforms.validators.DataRequired()])

    def validate_login(self, field):
        """Validate that login does not already exist."""
        if (
            db.session.query(models.AdminUser)
            .filter_by(login=self.login.data).count() > 0
        ):
            raise wtforms.validators.ValidationError('Duplicate username')


class CreateAdminUser(flask_script.Command):
    """Flask-Script command to create a new Admin user."""

    def run(self):
        """Create a new Admin user."""
        login = None
        password = None
        email = None

        while login is None:
            login = input('Login name: ')

        while email is None:
            email = input('Email: ')

        while password is None:
            password = getpass.getpass()
            password2 = getpass.getpass('Password (again): ')
            if password != password2:
                print("Error: Your passwords didn't match.")
                password = None
                # Don't validate passwords that don't match.
                continue

            if password.strip() == '':
                print("Error: Blank passwords aren't allowed.")
                password = None
                # Don't validate blank passwords.
                continue

        form = RegistrationForm(login=login, email=email,
                                password=password)
        if form.validate():
            user = models.AdminUser()

            form.populate_obj(user)
            # we hash the users password to avoid saving it as plaintext in the
            # db.
            user.password = werkzeug.generate_password_hash(form.password.data)
            user.active = True

            db.session.add(user)
            db.session.commit()
        else:
            print(form.errors)



import sqlalchemy as sa

from pg_discuss import ext
from pg_discuss import tables
from pg_discuss.db import db


class ArchiveCommentVersionsExt(ext.OnPostCommentUpdate,
                                ext.AddCommentFilterPredicate):
    """Extension to archive comment versions.

    When a comment has been edited through the HTTP API, a version is saved
    as another Comment record. Old versions always have a foreign key to the
    most recent version of the comment in `version_of_id`.
    """

    def on_post_comment_update(self, old_comment, new_comment, **extras):
        # "Archive" the old version of the comment by re-inserting it, with a
        # new pk.  Set `archived` to True.
        t = tables.comment
        old_comment['version_of_id'] = old_comment['id']
        del old_comment['id']
        old_comment['custom_json']['archived'] = True
        stmt = (
            t.insert()
            .values(**old_comment)
        )
        db.engine.execute(stmt)

    def add_comment_filter_predicate(self, **extras):
        """Return a filter clause to exclude archived comments from the default
        fetch.
        """
        t = tables.comment
        return t.c.custom_json['archived'].cast(sa.Boolean).isnot(True)



"""
Forms to validate new comments and comment edits. These forms will execute
any hooks associated with `ValidateComment` extensions.
"""
from voluptuous import (
    Schema,
    Required,
    All,
    Any,
    Invalid,
)

from . import _compat
from . import ext
from . import queries


def validate_parent(parent_id):
    """Validate that the parent_id associated with a new comment corresponds to
    a valid parent comment.
    """
    if parent_id is not None and not queries.validate_parent_exists(parent_id):
        raise Invalid('parent does not exist')
    return parent_id


def exec_comment_validators(action):
    """Validate comment with all configured validators.

    Validation Exceptions raised within validator functions will be propagated
    up and processing will be stopped.

    If no validator raises an Exception, the result of the last validator
    will be returned.

    The `action` string is supplied to each extension. Possible actions are:
    'create', 'edit'.
    """

    def validator(comment):
        """Validate the comment against all configured ValidateComment
        extensions.
        """
        result = ext.exec_hooks(ext.ValidateComment, comment, action)
        if len(result) > 0:
            return result[-1]
        # If no configured validators, just return original comment.
        else:
            return comment
    return validator


def validate_new_comment(new_comment):
    """Validate a new comment."""
    new_comment_schema = All(
        Schema({
            'parent_id': All(Any(int, None), validate_parent),
            Required('text'): _compat.text_type,
            Required('custom_json'): dict,
            Required('identity_id'): Any(int, None),
        }),
        exec_comment_validators(action='create'),
    )

    return new_comment_schema(new_comment)


def validate_comment_edit(comment_edit):
    """Validate a commment edit. This differs from validating a new comment in
    that `parent_id` is not allowed to be included - the parent cannot be
    changed by editing. Also, the `identity_id` may not be changed.
    """
    comment_edit_schema = All(
        Schema({
            Required('text'): _compat.text_type,
            Required('custom_json'): dict,
        }),
        exec_comment_validators(action='edit'),
    )

    return comment_edit_schema(comment_edit)



"""Add unique index for voting

Revision ID: 18d88bc2c83
Revises: 
Create Date: 2015-09-27 21:17:27.003073

"""

# revision identifiers, used by Alembic.
revision = '18d88bc2c83'
down_revision = None

from alembic import op

def upgrade():
    op.get_bind().execute('''
CREATE UNIQUE INDEX _voting_uc
ON identity_comment (identity_id, comment_id)
WHERE (rel_type = 'upvote' OR rel_type='downvote');
''')


def downgrade():
    op.get_bind().execute('''
DROP INDEX _voting_uc
''')



"""Extension base classes.

There are two types of extensions base clases: the Application Extension base
and Hook bases. Base classes can be used as mixins, so that it is possible for
an extensions to implement more than one behavior defined by a base class. As
an extreme example, it is possible to make an extension that implements the
Application Extension base and all Hook bases by subclassing all of the bases
in this module.

The Application Extension base class allows an extension to modify the Flask
app upon initialization, for example, to add routes or request/response
middleware.

The Hook base classes allow an extension to implement a method that is called
when a certain event occurs. To allow Hook APIs to send additional parameters
in the future, all Hook methods must take a `**extras` argument.
"""
import abc
import inspect
import six
import sys

from flask import current_app
import sqlalchemy as sa


class PluginLoadError(Exception):
    pass


class GenericExtBase(object):
    """
    Provide a default class with an __init__ that accepts an app argument.

    To be subclassed by both extension and driver ABCs. This allows
    all plugins to read the configuration from the app.
    """
    def __init__(self, app=None):
        self.app = app


# Driver ABCs
@six.add_metaclass(abc.ABCMeta)
class IdentityPolicy(GenericExtBase):
    """Policy to retrieve an Identity from an identity store, and
    remember/forget the identity (eg, in the browser session).

    The identity may be tied only to a browser session, or may be tied
    to an authenticated user associated with an authentication provider.
    """

    @abc.abstractmethod
    def get_identity(self, request, **extras):
        """Get the identity record associated with the request.
        """

    @abc.abstractmethod
    def forget(self, request, **extras):
        """Forget the identity, if remembered, on subsequent requests.
        """

    @abc.abstractmethod
    def remember(self, request, **extras):
        """Remember the identity for subsequent requests.
        """


@six.add_metaclass(abc.ABCMeta)
class CommentRenderer(GenericExtBase):
    """Driver to render comment text.
    """

    @abc.abstractmethod
    def render(self, text, **extras):
        """Render raw text into another format for display.
        """


# Extension ABCs
@six.add_metaclass(abc.ABCMeta)
class AppExtBase(GenericExtBase):
    """Base class for generic Flask app extensions.

    Such extensions will perform app initialization, possibly
    adding new views.
    """
    @abc.abstractmethod
    def init_app(self, app):
        """Perform app initialization."""


@six.add_metaclass(abc.ABCMeta)
class ValidateComment(GenericExtBase):
    """Mixin class for extensions that validate comment data. Validators
    may mutate the comment data into a valid form.

    Receives an `action` parameter which is either `create` or `edit`.
    """
    @abc.abstractmethod
    def validate_comment(self, comment, action, **extras):
        """Validate a comment dictionary.
        """
    hook_method = validate_comment.__name__


@six.add_metaclass(abc.ABCMeta)
class OnPreCommentInsert(GenericExtBase):
    """Mixin class for extensions that modify the insert statement for new
    comments.
    """
    @abc.abstractmethod
    def on_pre_comment_insert(self, new_comment, **extras):
        """Modify the comment object before inserting.
        """
    hook_method = on_pre_comment_insert.__name__


@six.add_metaclass(abc.ABCMeta)
class OnPostCommentInsert(GenericExtBase):
    """Mixin class for extensions that perform some action with the result of
    an insert.
    """
    @abc.abstractmethod
    def on_post_comment_insert(self, new_comment, **extras):
        """Perform some action with the result of an insert.
        """
    hook_method = on_post_comment_insert.__name__


@six.add_metaclass(abc.ABCMeta)
class OnPreCommentUpdate(GenericExtBase):
    """Mixin class for extensions that modify the update statement for comment
    updates.
    """
    @abc.abstractmethod
    def on_pre_comment_update(self, old_comment, comment_edit, **extras):
        """Modify the update statement for comment updates.
        """
    hook_method = on_pre_comment_update.__name__


@six.add_metaclass(abc.ABCMeta)
class OnPostCommentUpdate(GenericExtBase):
    """Mixin class for extensions that perform some action with the result of
    an update.
    """
    @abc.abstractmethod
    def on_post_comment_update(self, old_comment, new_comment, **extras):
        """Perform some action with the result of an insert.
        """
    hook_method = on_post_comment_update.__name__


@six.add_metaclass(abc.ABCMeta)
class AddCommentFilterPredicate(GenericExtBase):
    """Mixin class for extensions that add a filter predicate to the comment
    fetch.
    """
    @abc.abstractmethod
    def add_comment_filter_predicate(self, **extras):
        """Returns a predicate for the where clause for comment fetches.
        Will be joined with other predicates using AND.
        """
    hook_method = add_comment_filter_predicate.__name__


@six.add_metaclass(abc.ABCMeta)
class OnPreCommentSerialize(GenericExtBase):
    """Mixin class for extensions that want to add fields to the serialized
    comment.
    """
    @abc.abstractmethod
    def on_pre_comment_serialize(self, raw_comment, client_comment, **extras):
        """Add fields to the comment representation to be serialized,
        `client_comment`, from the dictionary representing the raw database
        row, `raw_comment`.
        """
    hook_method = on_pre_comment_serialize.__name__


@six.add_metaclass(abc.ABCMeta)
class OnPreThreadSerialize(GenericExtBase):
    """Mixin class for extensions that want to add fields to the serialized
    thread.
    """
    @abc.abstractmethod
    def on_pre_thread_serialize(self, raw_thread, comment_seq, client_thread,
                                **extras):
        """Add fields to the thread representation to be serialized,
        `client_thread`, from the dictionary representing the raw database row,
        `raw_thread`.
        """
    hook_method = on_pre_thread_serialize.__name__


@six.add_metaclass(abc.ABCMeta)
class OnNewCommentResponse(GenericExtBase):
    """Mixin class for extensions that want to process the response of the new
    comment view.

    A common case is to set cookies or other headers using the value of the
    newly created comment.
    """
    @abc.abstractmethod
    def on_new_comment_response(self, resp, raw_comment, client_comment,
                                **extras):
        """Process the response of the new comment view. Both the 'raw' comment
        and the comment formatted for the client are given as arguments.
        """
    hook_method = on_new_comment_response.__name__


def hook_classes():
    """Return a list of all hook classes in the `ext` module."""
    return [cls for name, cls in inspect.getmembers(sys.modules[__name__])
            if inspect.isclass(cls) and hasattr(cls, 'hook_method')]


def get_hook_map(extensions, hook_classes):
    """Create a map of hook classes to extensions that subclass them.
    This is a speedup to avoid having to do `isinstance` checks
    on extensions for every hook invocation.
    """
    hook_map = {}
    for cls in hook_classes:
        hook_map[cls] = [e.obj for e in extensions if isinstance(e.obj, cls)]
    return hook_map


# Extension utility functions
def exec_hooks(ext_class, *args, **kwargs):
    """Execute the hook function associated with the extension mixin class.
    Note that this allows for extensions to subclass multiple hook mixins.

    Returns the list of results obtained from calling all relevant
    extensions in order. Similar to stevedore's `ExtensionManager.map()`,
    however, we only call on extensions that are an instance of the base class
    given.
    """

    def execute_hook(ext_obj, ext_class, *args, **kwargs):
        """Invoke the hook associated with a given extension class on the
        extension object."""
        return getattr(ext_obj, ext_class.hook_method)(*args, **kwargs)

    results = []
    for ext_obj in current_app.hook_map[ext_class]:
        results.append(execute_hook(ext_obj, ext_class, *args, **kwargs))

    return results


def exec_init_app(app):
    """Execute the init_app for any instance of AppExtBase.
    """

    for ext in app.ext_mgr.extensions:
        if isinstance(ext.obj, AppExtBase):
            ext.obj.init_app(app)


def fail_on_ext_load(manager, entrypoint, exception):
    """Function to be used as `on_load_failure_callback` for stevedore.

    Raise an exception that includes the traceback from the failed
    plugin load.
    """
    import traceback
    msg = (
        'Error loading plugin {0}\n'
        'Plugin load error traceback:\n{1}'
        .format(entrypoint, traceback.format_exc()))
    raise PluginLoadError(msg)


def exec_filter_hooks(ext_class, stmt, *args, **kwargs):
    """Execute query filter hooks and collect all the returned predicates
    in an AND expression to be used in the WHERE clause.
    """
    predicates = exec_hooks(ext_class, *args, **kwargs)
    stmt = stmt.where(sa.and_(*predicates))
    return stmt



import flask
from voluptuous import Schema

from pg_discuss import _compat
from pg_discuss import ext

#: Allow the author field to be changed by editing a comment.
CAPTURE_AUTHOR_ALLOW_EDIT = False


class CaptureAuthor(ext.AppExtBase, ext.ValidateComment,
                    ext.OnPreCommentSerialize):
    """Extension to capture and persist the author field from the HTTP API.

    Enables a setting, `CAPTURE_AUTHOR_ALLOW_EDIT`, which when True,
    allows the author field to be edited by the Identity which posted.
    """

    def init_app(self, app):
        self._app = app
        app.config.setdefault('CAPTURE_AUTHOR_ALLOW_EDIT',
                              CAPTURE_AUTHOR_ALLOW_EDIT)

    def validate_comment(self, comment, action, **extras):
        allow_edit = self._app.config['CAPTURE_AUTHOR_ALLOW_EDIT']
        if action == 'create' or allow_edit:
            author = flask.request.get_json().get('author')
            if author:
                comment['custom_json']['author'] = (
                    Schema(_compat.text_type)(author)
                )
        return comment

    def on_pre_comment_serialize(self, raw_comment, client_comment, **extras):
        if 'author' in raw_comment['custom_json']:
            client_comment['author'] = raw_comment['custom_json']['author']



import flask
from voluptuous import (
    All,
    Schema,
    Url,
)

from pg_discuss import _compat
from pg_discuss import ext


class CaptureWebsite(ext.ValidateComment, ext.OnPreCommentSerialize):
    """Extension to capture and persist the website from the HTTP API."""
    def validate_comment(self, comment, action, **extras):
        website = flask.request.get_json().get('website')
        if website:
            form = Schema(All(_compat.text_type, Url()))
            url = form(website)
            comment['custom_json']['website'] = normalize_url(url)
        return comment

    def on_pre_comment_serialize(self, raw_comment, client_comment, **extras):
        if 'website' in raw_comment['custom_json']:
            client_comment['website'] = raw_comment['custom_json']['website']


def normalize_url(url):
    if not url.startswith(("http://", "https://")):
        return "http://" + url
    return url



from .. import ext


class NullIdentityPolicy(ext.IdentityPolicy):
    """An `IdentityPolicy` that never returns an identity."""
    def remember(self, request, identity_id, **extras):
        pass

    def get_identity(self, request, **extras):
        return None

    def forget(self, request, **extras):
        pass



"""SQLAlchemy `Table` objects for database tables.

The SQLAlchemy Expression Language is preferred for CRUD operations, so
shorter aliases are created for the table objects are created here.
"""
from . import models


thread = models.Thread.__table__
comment = models.Comment.__table__
identity = models.Identity.__table__
identity_comment = models.IdentityComment.__table__
admin_user = models.AdminUser.__table__



import flask
from flask_admin.actions import action
from flask_admin.babel import gettext, ngettext, lazy_gettext
from flask_admin.base import expose
from flask_admin.contrib.sqla import filters
import sqlalchemy as sa

from . import admin
from pg_discuss import ext
from pg_discuss import models
from pg_discuss import tables
from pg_discuss.db import db

#: Enable moderation by setting new posts to `pending`.
MODERATION_ACTIVE = True


class Moderate(admin.PrettyComment):
    pass


class ModerationExt(ext.AppExtBase, ext.AddCommentFilterPredicate,
    ext.OnPreCommentInsert):
    """Extension that requires comments to be approved by an admin user
    through the admin interface before publishing. Once this plugin is enabled,
    it can be deactivated by setting `MODERATION_ACTIVE` to False.
    Removing the plugin entirely will cause pending/rejected comments to
    appear, thus is not recommended unless some cleanup is done in the
    database.
    """
    def __init__(self, app):
        self.app = app
        app.config.setdefault('MODERATION_ACTIVE', MODERATION_ACTIVE)

    def init_app(self, app):
        app.admin.add_view(CommentAdminWithModeration(
            Moderate,
            models.db.session,
            name='Moderation',
            endpoint='moderation',
        ))

    def add_comment_filter_predicate(self, **extras):
        """Return a filter clause to exclude unapproved comments from the
        default fetch.
        """
        t = tables.comment
        return sa.or_(
            t.c.custom_json['mod_mode'].cast(sa.Text) == 'approved',
            sa.not_(t.c.custom_json.has_key('mod_mode'))
        )

    def on_pre_comment_insert(self, new_comment, **extras):
        """Set `mod_mode` to `pending`, unless `MODERATION_ACTIVE` is False.
        """
        if self.app.config['MODERATION_ACTIVE']:
            new_comment['custom_json']['mod_mode'] = 'pending'


class PendingApprovalFilter(filters.BaseSQLAFilter):
    def apply(self, query, value, alias=None):
        t = tables.comment
        return query.filter(t.c.custom_json['mod_mode'].cast(sa.Text) == value)

    def operation(self):
        return lazy_gettext('is')


class CommentAdminWithModeration(admin.AuthenticatedModelView):

    column_list = ('identity', 'thread', 'text')
    can_delete = False
    can_create = False
    column_filters = [PendingApprovalFilter(
        models.Comment.custom_json,
        'Approval Status',
        [('pending', 'Pending'),
         ('approved', 'Approved'),
         ('rejected', 'Rejected')])]

    @expose('/')
    def index_view(self):
        """Redirect to default 'Pending' filter if no query args."""
        if not flask.request.args:
            return flask.redirect(flask.request.url + '?flt1_0=pending')
        else:
            return super(CommentAdminWithModeration, self).index_view()

    @action('approve',
            lazy_gettext('Approve'),
            lazy_gettext('Are you sure you want to approve selected records?'))
    def action_approve(self, ids):
        self.set_approval(ids, 'approve')

    @action('reject',
            lazy_gettext('Reject'),
            lazy_gettext('Are you sure you want to reject selected records?'))
    def action_reject(self, ids):
        self.set_approval(ids, 'reject')

    def set_approval(self, ids, action):
        if action == 'approve':
            jsonb_val = 'approved'
            action_text = 'approved'
        elif action == 'reject':
            jsonb_val = 'rejected'
            action_text = 'rejected'

        try:
            t = tables.comment
            mod_mode = sa.func.jsonb_set(
                sa.text('custom_json'),
                sa.text("'{mod_mode}'"),
                sa.text('\'"{}"\''.format(jsonb_val))
            )
            stmt = (
                t.update()
                .where(t.c.id.in_(ids))
                .values(custom_json=mod_mode)
                .returning(t.c.id)
            )
            result = db.engine.execute(stmt).fetchall()
            count = len(result)

            flask.flash(ngettext(
                'Comment was successfully {}.'
                .format(action_text),
                '%(count)s comments were successfully {}.'
                .format(action_text),
                count,
                count=count))
        except Exception as ex:
            if not self.handle_view_exception(ex):
                raise

            flask.flash(gettext('Failed to {} records. %(error)s'
                                .format(action_text),
                                error=str(ex)), 'error')



"""Base configuration and configuration helper functions.

Defaults are appropriate for production use.

All connection parameters and secrets should be read from the environment.
"""
import os
import logging

#: Python recursion limit. This determines at what level of comment nesting
#: the JSON encoder will hit "RuntimeError: maximum recursion depth exceeded".
PYTHON_RECURSION_LIMIT = 30000

#: Flask-SQLAlchemy can track object modifications. Disable because of
#: memory overhead.
SQLALCHEMY_TRACK_MODIFICATIONS = False

# Extension settings
#: Configuration parameters of the form `ENABLE_EXT_*` are special:
#: the set of these parameters are parsed into a list of extension entrypoints
#: to be enabled.  The part of the string after the prefix `ENABLE_EXT_` is
#: taken to be the name of an extension entrypoint in all caps. All entrypoints
#: are required to have lowercase names.
#:
#: Example: if `ENABLE_EXT_FOO=True`, the extension with entrypoint name `foo`
#: will be enabled.
#:
#: The purpose of these "magic" config parameters is:
#:
#: - to allow atomic enablement of extensions without knowing apriori all of
#:   the entrypoint names.
#: - to avoid importing extensions that are not enabled. It is possible for
#:   extensions to "self-enable" ala stevedore's `EnabledExtensionManager`,
#:   which requires that extensions be imported before checking to see if they
#:   are enabled.
#: - to avoid having to list all extensions in a comma-separated string, which
#:   complicates atomic enabling and disabling of extensions.
#
# See the `get_enabled_extensions` in this module for parsing details.
ENABLE_EXT_BLESSED_ADMIN = True
#:
ENABLE_EXT_BLESSED_CSRF_TOKEN = True
#:
ENABLE_EXT_BLESSED_ROUTE_LIST = True
#:
ENABLE_EXT_BLESSED_ARCHIVE_COMMENT_VERSIONS = True
#:
ENABLE_EXT_BLESSED_VALIDATE_COMMENT_LEN = True
#:
ENABLE_EXT_BLESSED_ISSO_CLIENT_SHIM = True
#:
ENABLE_EXT_BLESSED_CORS = True
#:
ENABLE_EXT_BLESSED_CAPTURE_AUTHOR = True
#:
ENABLE_EXT_BLESSED_CAPTURE_EMAIL = True
#:
ENABLE_EXT_BLESSED_CAPTURE_WEBSITE = True
#:
ENABLE_EXT_BLESSED_CAPTURE_REMOTE_ADDR = True
#:
ENABLE_EXT_BLESSED_VOTING = True
#:
ENABLE_EXT_BLESSED_MODERATION = False
#:
ENABLE_EXT_BLESSED_MOD_EMAIL = False
#:
ENABLE_EXT_BLESSED_PROFILER = False
#:
ENABLE_EXT_BLESSED_DOZER = False

#: Optional: Order extensions using comma-separated list of extension names.
#: It is generally discouraged to write order-dependent extensions, but
#: is required in some cases.
#: It is not necessary to list all extensions, only those that are required
#: to be ordered. See `config:sorted_ext_names` for sort logic details.
EXT_ORDER = (
    'blessed_voting,'
    'blessed_admin,'
    'blessed_moderation,'
    'blessed_isso_client_shim,'
)

# Driver settings
#: Comment renderer driver to use (as a setuptools entrypoint name)
DRIVER_COMMENT_RENDERER = 'blessed_markdown_renderer'
#: JSON encoder driver to use (as a setuptools entrypoint name)
DRIVER_JSON_ENCODER = 'blessed_unix_time_json_encoder'
#: Identity driver to use (as a setuptools entrypoint name)
DRIVER_IDENTITY_POLICY = 'blessed_auth_tkt_identity_policy'

# Session settings
#: Expiration of a permanent sesison in seconds.
PERMANENT_SESSION_LIFETIME = 3600
#: Default cookies to use `Secure` flag to avoid leaking unique machine
#: identifiers over the network. Non-HTTPS deployments must explicitly disable.
#: https://www.owasp.org/index.php/SecureFlag
SESSION_COOKIE_SECURE = True
#: Default cookies to use `HttpOnly` flag to deny client side scripts access
#: to the cookie. Defaults to True in Flask, but we want to be explicit.
#: https://www.owasp.org/index.php/HttpOnly
SESSION_COOKIE_HTTP_ONLY = True

# Log settings
#: Log level
LOGLEVEL = logging.INFO

#: List of parameters which should not be logged. This should list any
#: configuration parameters which include secrets, such as SECRET_KEY,
#: or the database connection string which may have a password.
DO_NOT_LOG_VARS = [
    'SECRET_KEY',
    'SQLALCHEMY_DATABASE_URI',
]

# Connection parameters and secrets
#: Database connection string.
SQLALCHEMY_DATABASE_URI = os.environ.get('DATABASE_URL')
#: Secret key used for cookie signing.
SECRET_KEY = os.environ.get('SECRET_KEY')
#: Name and port number of the server.
#: SERVER_NAME is required by some extensions that need to build absolute urls
#: in the context of emails, etc.
SERVER_NAME = os.environ.get('SERVER_NAME')

#: List of allowed origins (origins where widget will be embedded)
#: http://www.w3.org/TR/cors/
#: Attempt to read CORS_ORIGIN from environment as a comma-separated list of
#: origins.
CORS_ORIGINS = []
cors_origins_from_env = os.environ.get('CORS_ORIGINS')
if cors_origins_from_env:
    CORS_ORIGINS = cors_origins_from_env.split(',')

#: List of headers that client is allowed to read.
CORS_EXPOSE_HEADERS = ['Date', 'X-Set-Cookie', 'X-CSRF-Token']


def get_enabled_extensions(config):
    """Get the list of extension entrypoint names where `ENABLE_EXT_*` is True.
    """
    prefix = 'ENABLE_EXT_'

    enable_ext_vars = [k for k in config.keys() if k.startswith(prefix)]

    ext_names = [k[len(prefix):].lower() for k in enable_ext_vars if config[k]]

    return sorted_ext_names(config, ext_names)


def sorted_ext_names(config, ext_names):
    """Sort extensions if `EXT_ORDER` is specified. Extensions not listed in
    `EXT_ORDER` will be appended to the end of the list in an arbitrary order.
    """
    ext_order = [e.strip() for e in config['EXT_ORDER'].split(',')]

    def sort_key(ext_name):
        try:
            return ext_order.index(ext_name)
        except ValueError:
            # If `ext_name` is not in the sort list, it has a sort value of
            # positive infinity (last).
            return float('inf')

    return sorted(ext_names, key=sort_key)



"""Set up the SQLAlchemy database engine.
"""
import flask_sqlalchemy


class PgAlchemy(flask_sqlalchemy.SQLAlchemy):
    """Custom subclass of the SQLAlchemy extension that sets the
    connection timezone to UTC. The backend should handle timestamps entirely
    in UTC, with timezone adjustment performed on the client side.
    """
    def apply_driver_hacks(self, app, info, options):
        options['connect_args'] = {"options": "-c timezone=utc"}
        options['isolation_level'] = 'AUTOCOMMIT'

db = PgAlchemy()



from werkzeug.contrib.profiler import ProfilerMiddleware

from pg_discuss import ext

#: Profile report restrictions.
#: See: https://docs.python.org/2/library/profile.html#pstats.Stats.print_stats
PROFILER_RESTRICTIONS = [20]
#: Directory to save profiler results. Results can be interpreted by a
#: graphical program such as SnakeViz (highly recommended):
#: https://jiffyclub.github.io/snakeviz/
PROFILER_DIR = None


class ProfilerExt(ext.AppExtBase):
    """Extension to enable the werkzeug `ProfilerMiddleware`.
    See: http://werkzeug.pocoo.org/docs/0.10/contrib/profiler/
    """
    def init_app(self, app):
        app.config.setdefault('PROFILER_RESTRICTIONS', PROFILER_RESTRICTIONS)
        app.config.setdefault('PROFILER_DIR', PROFILER_DIR)
        app.config['PROFILE'] = True
        restrictions = app.config['PROFILER_RESTRICTIONS']
        profile_dir = app.config['PROFILER_DIR']
        app.wsgi_app = ProfilerMiddleware(
            app.wsgi_app,
            restrictions=restrictions,
            profile_dir=profile_dir)



"""Generic utility functions for pg-discuss."""


class MergeConflict(Exception):
    pass


def merge_fail_on_conflict(dict1, dict2):
    """Merge two dictionaries, failing on conflict."""
    for key in dict2.keys():
        if key in dict1:
            raise MergeConflict("key {} already exists".format(key))
        else:
            return dict(dict1, **dict2)



"""
The core schema is intended to be as simple as possible, defining only the
attributes that are absolutely necessary for minimal, relational comment
objects.

Thread schema:

 - id: Primary key.
 - client_id: String used as unique identifier by client.
 - custom_json

`client_id` must have a value which is *immutable*.  Client uses this to
create the initial thread and fetch comments associated with a thread.

URI, title, slug are *not* recommended because these values may change (page
may move to different domain, title/slug may be edited).

Recommended: For many static site generators, a client_id can be defined in
custom metadata. The client_id can be set to a UUID generated at the time of
initial article creation (using a text editor plugin, or manually using a CLI
tool such as uuidgen). The static site generator can then render this in to an
HTML data attribute for extraction by the client.

Comment schema:

 - id: Primary key.
 - thread_id: Foreign key to thread.
 - parent_id: Foreign key to parent comment.
 - version_of_id: Foreign key to the comment of which this comment is a
   version. Used by extensions that preserve comment versions.
 - created: Creation timestamp.
 - modified: Modified timestamp.
 - text: The comment text itself.
 - custom_json

custom_json is intended for custom attributes. As a JSON blob, the schema is
flexible and can be extended without requiring database migrations or changes
to the core code. Some possible uses:

 - email
 - website
 - archived flag
 - deleted flag
 - moderated flag
"""
from sqlalchemy import (
    Boolean,
    Column,
    DateTime,
    ForeignKey,
    Integer,
    String,
    text,
)
from sqlalchemy.orm import (
    backref,
    relationship,
)
from sqlalchemy.dialects.postgresql import JSONB

from .db import db


class Thread(db.Model):
    """Thread model."""
    id = Column(
        Integer,
        primary_key=True)
    client_id = Column(
        String,
        unique=True)
    created = Column(
        DateTime(timezone=True),
        server_default=text('NOW()'),
        nullable=False)
    custom_json = Column(
        JSONB,
        server_default='{}',
        nullable=False)

    def __unicode__(self):
        return self.client_id
    __str__ = __unicode__


class Comment(db.Model):
    """Comment model."""
    id = Column(
        Integer,
        primary_key=True,
        nullable=False)
    identity_id = Column(
        Integer,
        ForeignKey('identity.id'),
        nullable=True)
    thread_id = Column(
        Integer,
        ForeignKey('thread.id'),
        nullable=False)
    parent_id = Column(
        Integer,
        ForeignKey('comment.id'))
    version_of_id = Column(
        Integer,
        ForeignKey('comment.id'),
        nullable=True)
    created = Column(
        DateTime(timezone=True),
        server_default=text('NOW()'),
        nullable=False)
    modified = Column(
        DateTime(timezone=True),
        server_default=text('NOW()'),
        nullable=False)
    text = Column(
        String,
        nullable=False)
    custom_json = Column(
        JSONB,
        server_default='{}',
        nullable=False)
    identity = relationship(
        'Identity',
        backref=backref('comments', cascade='all, delete-orphan'))
    thread = relationship(
        'Thread',
        backref=backref('comments', cascade='all, delete-orphan'))

    def __unicode__(self):
        return '<{}> {}'.format(self.identity, self.text[:40])
    __str__ = __unicode__


class Identity(db.Model):
    """Identity model."""
    id = Column(
        Integer,
        primary_key=True,
        nullable=False)
    created = Column(
        DateTime(timezone=True),
        server_default=text('NOW()'),
        nullable=False)
    custom_json = Column(
        JSONB,
        server_default='{}',
        nullable=False)

    def __unicode__(self):
        return '({}) {}'.format(self.id, self.custom_json)
    __str__ = __unicode__


class IdentityComment(db.Model):
    """Association table between Identities and Comments, which may be used to
    by extensions to represent many-to-many relationships such as
    "upvotes"/"downvotes".

    This model is not utilized by the pg-discuss core.
    """
    __tablename__ = 'identity_comment'
    id = Column(
        Integer,
        primary_key=True,
        nullable=False)
    identity_id = Column(
        Integer,
        ForeignKey('identity.id'),
        nullable=False)
    comment_id = Column(
        Integer,
        ForeignKey('comment.id'),
        nullable=False)
    rel_type = Column(
        String,
        nullable=False)
    created = Column(
        DateTime(timezone=True),
        server_default=text('NOW()'),
        nullable=False)
    custom_json = Column(
        JSONB,
        server_default='{}',
        nullable=False)
    identity = relationship(
        "Identity",
        backref=backref(
            'identity_comment',
            cascade='all, delete-orphan'))
    comment = relationship(
        "Comment",
        backref=backref(
            'identity_comment',
            cascade='all, delete-orphan'))

    def __unicode__(self):
        return '<{}> {} <{}>'.format(self.identity, self.rel_type,
                                     self.comment)
    __str__ = __unicode__


class AdminUser(db.Model):
    """Model for managing admin user authentication.

    This model is not for general user authentication. If traditional
    username/password authentication is desired, it must be implemented as
    an `IdentityPolicy` with its own model.
    """
    __tablename__ = 'admin_user'
    id = Column(
        Integer,
        primary_key=True,
        nullable=False)
    login = Column(
        String,
        unique=True,
        nullable=False)
    email = Column(
        String,
        nullable=False)
    active = Column(
        Boolean,
        nullable=False)
    password = Column(
        String,
        nullable=False)

    # Flask-Login integration
    def is_authenticated(self):
        """Is this user authenticated? Always True."""
        return True

    def is_active(self):
        """Is this user active?"""
        return self.active

    def is_anonymous(self):
        """Is this user anonymous? Always False."""
        return False

    def get_id(self):
        return self.id

    # Required for administrative interface
    def __unicode__(self):
        return self.username
    __str__ = __unicode__



import markupsafe

from .. import ext


class EscapingRenderer(ext.CommentRenderer):
    """Renderer driver that escapes using HTML entity encoding.

    This neutralizes all markup contained in the comment text.
    """

    def render(self, text, **extras):
        """Render to HMTL entity-encoded text.
        """
        return markupsafe.escape(text)



import re

import flask
from voluptuous import (
    All,
    Invalid,
    Schema,
)

from pg_discuss import _compat
from pg_discuss import ext


class CaptureEmail(ext.ValidateComment):
    """Extension to capture and persist the email field from the HTTP API.
    """
    def validate_comment(self, comment, action, **extras):
        email = flask.request.get_json().get('email')
        if email:
            form = Schema(All(_compat.text_type, validate_email))
            comment['custom_json']['email'] = form(email)
        return comment


def validate_email(v):
    if re.match("[\w\.\-\+]*@[\w\.\-]*\.\w+", v):
        return v
    else:
        raise Invalid("email is not in recognized format: {0}".format(v))



import operator

import misaka

from pg_discuss import _compat
from pg_discuss import ext

#: Markdown rendering flags. These flags have integer values: the default of
#: "2" is for the `HTML_ESCAPE` flag.
#: See: http://misaka.61924.nl/api/
MARKDOWN_RENDER_FLAGS = [misaka.HTML_ESCAPE, misaka.HTML_HARD_WRAP]
#: Markdown extension flags. These flags have integer values: the default of
#: "2" is for the `EXT_FENCED_CODE` extension.
#: See: http://misaka.61924.nl/api/
MARKDOWN_EXTENSION_FLAGS = [misaka.EXT_FENCED_CODE]


class MarkdownRenderer(ext.CommentRenderer):
    """
    Markdown render driver via Mikasa 2, a CFFI wrapper for the Hoedown C
    Markdown renderer.

    This renderer escapes any HTML that does not originate from the allowed
    Markup syntax.
    """

    def __init__(self, app=None):
        # TODO: Allow flags to be given as a list in a configuration
        # variable. Need to wait until Mikasa 2 releases with stable API.
        app.config.setdefault('MARKDOWN_RENDER_FLAGS',
                              MARKDOWN_RENDER_FLAGS)
        app.config.setdefault('MARKDOWN_EXTENSION_FLAGS',
                              MARKDOWN_EXTENSION_FLAGS)

        # Flags/extensions must be OR'd together to get an integer.
        render_flags_int = _compat.reduce(operator.or_,
                                          app.config['MARKDOWN_RENDER_FLAGS'])
        extensions_int = _compat.reduce(operator.or_,
                                        app.config['MARKDOWN_EXTENSION_FLAGS'])

        renderer = misaka.HtmlRenderer(flags=render_flags_int)
        self.markdown = misaka.Markdown(renderer, extensions=extensions_int)

    def render(self, text, **extras):
        """Render Markdown to HTML.
        """
        return self.markdown(text)



import codecs
import datetime
import functools

from flask import request
import pytz
import simplejson as json
import werkzeug
import werkzeug.security

from pg_discuss import ext


class IssoClientShim(ext.AppExtBase, ext.OnPreCommentSerialize,
                     ext.OnPreThreadSerialize, ext.OnPreCommentInsert,
                     ext.OnNewCommentResponse):
    """Extension to adapt JSON API to format used by Isso JavaScript client.

    If extensions that add edit/delete views are enabled, they must be loaded
    before this extension.
    """

    def init_app(self, app):
        # Disable all pretty-printing. Flask will not disable it since
        # `X-Requested-With: XMLHttpRequest` is not sent.
        app.config['JSONIFY_PRETTYPRINT_REGULAR'] = False

        views = app.view_functions

        # Exempt public read-only views from IdentityPolicy
        app.identity_policy_mgr.exempt(self.fetch_)
        app.identity_policy_mgr.exempt(self.count)

        app.route('/', methods=['GET'])(self.fetch_)
        app.route('/new', methods=['POST'])(self.new_)
        app.route('/id/<int:comment_id>', methods=['GET'])(views['view'])
        app.route('/count', methods=['POST'])(self.count)
        app.route('/id/<int:comment_id>', methods=['PUT'])(views['edit'])
        app.route('/id/<int:comment_id>', methods=['DELETE'])(views['delete'])
        app.route('/id/<int:comment_id>/like', methods=['POST'])(self.like_)
        app.route('/id/<int:comment_id>/dislike', methods=['POST'])(
            self.dislike_)

    def on_pre_comment_serialize(self, raw_comment, client_comment, **extras):
        # Change `parent_id` key to `parent`
        client_comment['parent'] = raw_comment.pop('parent_id')

        # Change `upvotes` to `likes`, `downvotes` to `dislikes`
        if 'upvotes' in client_comment:
            client_comment['likes'] = client_comment.pop('upvotes')
        if 'downvotes' in client_comment:
            client_comment['dislikes'] = client_comment.pop('downvotes')

        client_comment['hash'] = raw_comment['custom_json'].get('hash')

        # Set "mode" to 4 for deleted comments
        if raw_comment['custom_json'].get('deleted'):
            client_comment['mode'] = 4

    def on_pre_thread_serialize(self, raw_thread, comment_seq, client_thread,
                                **extras):
        # We interpret the client options slightly differently than the
        # Isso backend. `limit` is treated as a limit to the number of
        # direct descendants that will be returned for each node,
        # and `nested_limit` is treated as a limit to the depth of replies
        # returned.
        after = None
        try:
            # Because of a regression in Python 3, we must jump through
            # some hoops to correctly convert a Decimal string representing
            # a fractional Unix timestamp to a Datetime object.
            # See: https://bugs.python.org/issue23607
            # Naively using floats will result in comparison errors.
            after_str = request.args.get('after')
            if after_str:
                try:
                    seconds_str, ms_str = after_str.split('.')
                except ValueError:
                    seconds_str = after_str
                    ms_str = None

                after = datetime.datetime.fromtimestamp(
                    int(seconds_str),
                    tz=pytz.utc)
                if ms_str:
                    td = datetime.timedelta(microseconds=int(ms_str))
                    after = after + td
        except TypeError:
            pass

        try:
            reply_limit = int(request.args.get('limit'))
        except TypeError:
            reply_limit = None

        try:
            reply_depth_limit = int(request.args.get('nested_limit'))
        except TypeError:
            reply_depth_limit = None

        try:
            parent_id = int(request.args.get('parent'))
        except TypeError:
            parent_id = None

        # Change key to comment collection from "comments" to "replies"
        comment_tree = build_comment_tree(
            comment_seq=comment_seq,
            parent_id=parent_id,
            after=after,
            reply_limit=reply_limit,
            count_limit=None,
            reply_depth_limit=reply_depth_limit,
        )
        client_thread.update(comment_tree)
        del client_thread['comments']
        # Isso threads have a null `id` attribute
        client_thread['id'] = None

        # Add the count of all comments as `total_replies_in_thread`
        client_thread['total_replies_in_thread'] = client_thread['reply_count']

    def on_pre_comment_insert(self, new_comment, **extras):
        """Hash email, or remote_addr.

        If neither available, just skip hashing.
        """
        custom_json = new_comment['custom_json']
        val_to_hash = (
            custom_json.get('email')
            or custom_json.get('remote_addr')
        )
        if val_to_hash:
            custom_json['hash'] = hash(val_to_hash)

    def on_new_comment_response(self, resp, raw_comment, client_comment,
                                **extras):
        """Set a cookie to allow the client to edit/delete the comment.
        """
        # Set a cookie with a non-empty string value.
        # It does not matter to the client what the value is, just so long
        # as it is non-empty. Identitys are authenticated through another
        # means.
        cookie = functools.partial(
            werkzeug.dump_cookie,
            value='.',
            max_age=self.app.permanent_session_lifetime.total_seconds()
        )

        comment_id = raw_comment['id']

        resp.headers.add("Set-Cookie", cookie(str(comment_id)))
        resp.headers.add("X-Set-Cookie", cookie("isso-%i" % comment_id))

        return resp

    def new_(self):
        """Create a new comments.
        Use the `uri` paramters as the thread `client_id`. Note that
        the client should configure the `uri` parameter to be a unique client
        id, not a uri.
        """
        # Use the `uri` parameter as the thread `client_id`.
        thread_client_id = request.args.get('uri')

        # Set `parent_id` to parent.
        json_data = request.get_json()
        json_data['parent_id'] = json_data.get('parent')

        # Return response. Response will be further processed by
        # `on_new_comment_response` to set cookies.
        return self.app.view_functions['new'](thread_client_id)

    def fetch_(self):
        """Fetch the list of comments associated with the thread.
        Use the `uri` paramters as the thread `client_id`.
        """
        # Use the `uri` parameter as the thread `client_id`.
        thread_client_id = request.args.get('uri')
        # Render the landing page if there is no `uri` parameter.
        if not thread_client_id:
            return self.app.view_functions['landing']()
        resp = self.app.view_functions['fetch'](thread_client_id)
        # Set a `Date` HTTP header with the current datetime.
        resp.headers['Date'] = datetime.datetime.now().isoformat() + 'Z'
        return resp

    def count(self):
        """Not implemented, stubbed to satisfy client.
        """
        return '[]'

    def like_(self, comment_id):
        resp = self.app.view_functions['upvote'](comment_id)
        return rename_voting_keys(resp)

    def dislike_(self, comment_id):
        resp = self.app.view_functions['downvote'](comment_id)
        return rename_voting_keys(resp)


def rename_voting_keys(resp):
    d = json.loads(resp.get_data())
    d['likes'] = d.pop('upvotes')
    d['dislikes'] = d.pop('downvotes')
    resp.set_data(json.dumps(d))
    return resp


def build_comment_tree(comment_seq,
                       parent_id=None,
                       index=0,
                       after=None,
                       reply_limit=None,
                       count_limit=None,
                       reply_depth_limit=None):
    """Build the nested tree of comments, counting the number of replies to
    each comments as we go. Our goal is to have the subtree of comments
    that belong to the parent, starting at the specified index, with up
    to `count_limit` replies, sorted by chronological order, and no
    replies at a depth greater than `reply_depth_limit`. All this processing is
    expensive, but is a fairly intuitive way of selecting a subtree to allow
    for both incrementally loading long, flat threads as well as digging deeper
    into highly nested comments.

    We assume that the sequence of
    comments has been sorted by `created` time, which enables us to construct
    the tree by iterating over the list sequentially.

    Pagination:
    `parent` is the parent comment we want to retrieve nested comments for.
    If `parent` is None, then we assume we are retrieving the top-level
    comments.
    `index` is the index relative to the chronologically ordered set of
    comments to a parent (or, the top-level comments).
    """
    # Construct the full tree, starting from the parent.
    # Once we have the full tree starting from the parent, we can
    # sort/annotate/prune it however we like.
    comment_tree = construct_full_tree(comment_seq, parent_id)

    # Find index of the first comment created after `after`
    if after:
        for i, comment in enumerate(comment_tree['replies']):
            if comment['created'] > after:
                index = max(i, 0)
                break

    # We start by discarding all comments up to `index`.
    del comment_tree['replies'][:index]
    # We can now descend the tree and annotate each node with counts
    # of its descendants and later siblings+sibling descendants.
    annotate_counts(comment_tree)

    # We need to discard any nodes at a depth greater than `reply_depth_limit`.
    discard_beyond_depth_limit(comment_tree, depth_limit=reply_depth_limit)

    # Keep only the first `reply_limit` nodes for each subtree.
    # This ensures that there are not more than `reply_limit` replies
    # returned for any given comment.
    if reply_limit:
        keep_reply_limit(comment_tree, reply_limit)

    # Keep only the first `count_limit` nodes by chronology.
    # This puts an absolute limit on the number of comments returned.
    if count_limit:
        keep_count_limit(comment_tree, count_limit)

    return comment_tree


def construct_full_tree(comment_seq, parent_id=None):
    """Construct the full tree, given an ordered sequence of comments."""
    comment_tree = {'replies': []}
    comment_dict = {}

    for c in comment_seq:
        # Track which comment is the desired parent, if supplied.
        if parent_id and c['id'] == parent_id:
            parent = c

        c['replies'] = []
        # Index the comment into the dictionary by id.
        comment_dict[c['id']] = c
        # Increment the total comment count.

        # If no parent_id, must be a top-level comment.
        if not c['parent_id']:
            comment_tree['replies'].append(c)

        # Otherwise, this is a nested comment, and we append it to the replies
        # of the parent, using the map to look up the parent. The parent is
        # guaranteed to exist in the map, since our `comment_seq` array is
        # sorted by `created` time.
        else:
            comment_dict[c['parent_id']]['replies'].append(c)

    # Return the tree starting with the desired parent, if supplied.
    if parent_id:
        return parent
    # Else, return the complete tree.
    return comment_tree


def annotate_counts(node):
    """Recursive function to annotate each node with the count of all
    of it's descendants, as well as the count of all sibling comments plus
    their children, made after each node.
    """
    # If no replies, this is a leaf node. Stop and return 1.
    node['reply_count'] = 0
    if not node['replies']:
        return 1
    else:
        # Annotate descendants and sum counts.
        for r in node['replies']:
            node['reply_count'] += annotate_counts(r)
        # Once descendants are annotated with descendant counts,
        # annotate with the count of siblings and their children coming after
        # this node.
        after_count = 0
        for r in reversed(node['replies']):
            r['after_count'] = after_count
            after_count += r['reply_count'] + 1
        return node['reply_count'] + 1


def keep_reply_limit(node, reply_limit):
    """Keep only the first `count_limit` nodes by chronology."""
    def walk_tree_and_discard(n):
        # Discard any direct descendants beyond the `reply_limit`.
        if 'replies' in n:
            if len(n['replies']) > reply_limit:
                hidden_replies = n.get('hidden_replies', 0)
                c = n['replies'][reply_limit]
                num_discarded = c['after_count'] + c['reply_count'] + 1
                n['hidden_replies'] = hidden_replies + num_discarded

                del n['replies'][reply_limit:]

            # Recursively check the remaining replies.
            for r in n['replies']:
                walk_tree_and_discard(r)
    walk_tree_and_discard(node)


def keep_count_limit(node, count_limit):
    """Keep only the first `count_limit` nodes by chronology."""
    # First, we need to find the `created` time of the `count_limit`-th node.
    created_times = []

    def walk_tree_for_created(n):
        created = n.get('created')
        if created:
            created_times.append(n['created'])
        if 'replies' in n:
            for r in n['replies']:
                walk_tree_for_created(r)

    walk_tree_for_created(node)

    created_times.sort()

    keep_time = created_times[count_limit]

    def walk_tree_and_discard(n):
        # Discard any direct descendants that are greater than the time.
        if 'replies' in n:
            len_replies = len(n['replies'])
            num_discarded = 0
            for i in range(len_replies - 1, -1, -1):
                c = n['replies'][i]
                if c['created'] > keep_time:
                    num_discarded += c['after_count'] + c['reply_count'] + 1
                    del n['replies'][i]

            hidden_replies = n.get('hidden_replies', 0)
            n['hidden_replies'] = hidden_replies + num_discarded

        # Recursively check the remaining replies.
        for r in n['replies']:
            walk_tree_and_discard(r)
    walk_tree_and_discard(node)


def discard_beyond_depth_limit(node, depth_limit, cur_depth=0):
    """Discard nodes beyond the depth limit. The root is considered "level 0",
    and the first descendants are "level 1".
    """
    if cur_depth > depth_limit:
        del node['replies']
    else:
        for r in node['replies']:
            discard_beyond_depth_limit(r, depth_limit, cur_depth + 1)


def hash(val):
    salt = b"Eech7co8Ohloopo9Ol6baimi"
    hashed = werkzeug.security.pbkdf2_bin(
        val.encode('utf-8'), salt, 1000, 6, "sha1")
    return codecs.encode(hashed, "hex_codec").decode("utf-8")



"""Middleware to execute the configured `IdentityPolicy`.
"""
import flask


class IdentityPolicyManager(object):
    """Middleware to execute the configured IdentityPolicy.

    Before each request, calls the `get_identity` method of the
    configured `IdentityPolicy`. The `IdentityPolicy` may return an identity
    object, which is then assigned to the `flask.g` request global and the
    `remember` method of the `IdentityPolicy` method is invoked with the
    identity. If no identity is returned by the `IdentityPolicy`, request
    processing continues normally without the identity.
    """

    def __init__(self, app, identity_policy_cls):
        self.identity_policy = identity_policy_cls(app)
        self._exempt_views = []

        app.config.setdefault('IDENTITY_POLICY_EXEMPT_METHODS', ['OPTIONS'])

        @app.before_request
        def _auth_before_request():
            """Check if the request is exempted from the `IdentityPolicy`
            before invoking it.
            """
            if (
                flask.request.method in
                app.config['IDENTITY_POLICY_EXEMPT_METHODS']
            ):
                return
            if self._exempt_views:
                if not flask.request.endpoint:
                    return

                view = app.view_functions.get(flask.request.endpoint)
                if not view:
                    return

                dest = '%s.%s' % (view.__module__, view.__name__)
                if dest in self._exempt_views:
                    return
            return self.auth_before_request()

    def auth_before_request(self):
        """Get the identity object from the `IdentityPolicy`. If an identity is
        returned, set it on the `flask.g` request global and remember it.
        """
        identity = self.identity_policy.get_identity(flask.request)

        if identity:
            # Store the identity object on the `g` request global.
            flask.g.identity = identity

            # Remember the identity.
            self.identity_policy.remember(flask.request, identity['id'])

    def exempt(self, view):
        """Exclude a view from the IdentityPolicy middleware. Takes a view
        functions as the single argument.
        """
        view_location = '%s.%s' % (view.__module__, view.__name__)
        self._exempt_views.append(view_location)



"""Queries used by pg-discuss core and available to extensions."""
import sqlalchemy as sa
import sqlalchemy.dialects.postgresql

from . import ext
from . import tables
from . import utils
from .db import db


class CommentNotFoundError(Exception):
    pass


class ThreadNotFoundError(Exception):
    pass


class IdentityNotFoundError(Exception):
    pass


def fetch_thread_by_client_id(thread_client_id):
    """Fetch a thread object by thread_client_id from the database."""
    t = tables.thread
    stmt = t.select().where(t.c.client_id == thread_client_id)

    # TODO: Run on_pre_thread_fetch hooks
    # stmt = ext.exec_filter_hooks(ext.OnPreThreadFetch, stmt)

    result = db.engine.execute(stmt).first()

    if not result:
        return None
    thread = dict(result.items())

    return thread


def fetch_comment_by_id(comment_id):
    """Fetch a single comment object by id from the database."""
    t = tables.comment
    stmt = t.select().where(t.c.id == comment_id)

    # Run add_comment_filter_predicate hooks
    stmt = ext.exec_filter_hooks(ext.AddCommentFilterPredicate, stmt)

    result = db.engine.execute(stmt).first()
    if not result:
        raise CommentNotFoundError('Comment {0} not found'.format(comment_id))
    comment = dict(result.items())

    return comment


def fetch_comments_by_thread_client_id(thread_client_id):
    """Fetch a list of comments for the given thread's client_id from the
    database."""
    t_comment = tables.comment
    t_thread = tables.thread
    stmt = (
        sa.select(t_comment.c)
        .select_from(sa.join(t_comment, t_thread))
        .where(t_thread.c.client_id == thread_client_id)
        .order_by(sa.asc(t_comment.c.created))
    )

    # Run add_comment_filter_predicate hooks
    stmt = ext.exec_filter_hooks(ext.AddCommentFilterPredicate, stmt)

    result = db.engine.execute(stmt)
    # Very large result sets can cause a lot of memory allocation here
    # that CPython may not give back to the OS, due to a lack of compacting
    # GC. See: http://stackoverflow.com/a/5495318
    # However, assuming a worst case of 50K smallish rows, the memory for the
    # process will only roughly double (to about 100MB).
    # If comments could be sorted in the database, it would not be necessary
    # to fetch all the results and sort in-process. This has the drawback
    # of putting more load on the database server. If memory due to large
    # allocations for results is an issue, it is recommended to reload
    # workers after exceeding a memory limit via uwsgi's `reload-on-rss`.
    comments_seq = [dict(x) for x in result.fetchall()]
    result.close()

    return comments_seq


def insert_comment(new_comment):
    """Insert the `new_comment` object in to the database."""
    t = tables.comment
    stmt = (
        t.insert()
        .values(**new_comment)
        .returning(*list(t.c))
    )

    # Run on_pre_insert hooks
    ext.exec_hooks(ext.OnPreCommentInsert, new_comment)

    result = db.engine.execute(stmt).first()
    comment = dict(result.items())

    # Run on_post_insert hooks
    ext.exec_hooks(ext.OnPostCommentInsert, comment)

    return comment


def insert_thread(new_thread):
    """Insert the `new_thread` object in to the database."""
    t = tables.thread
    stmt = (
        t.insert()
        .values(**new_thread)
        .returning(*list(t.c))
    )

    # TODO: Run on_pre_thread_insert hooks
    # ext.exec_hooks(ext.OnPreThreadInsert, new_thread)

    result = db.engine.execute(stmt).first()
    thread = dict(result.items())

    # TODO: Run on_post_thread_insert hooks
    # ext.exec_hooks(ext.OnPostThreadInsert, thread)

    return thread


def insert_identity(new_identity=None):
    """Insert the a new identity object in to the database."""
    t = tables.identity
    stmt = (
        t.insert()
        .returning(*list(t.c))
    )
    if new_identity:
        stmt = stmt.values(**new_identity)

    # TODO: Run on_pre_identity_insert hooks
    # ext.exec_hooks(ext.OnPreIdentityInsert, new_identity)

    result = db.engine.execute(stmt).first()
    identity = dict(result.items())

    # TODO: Run on_post_identity_insert hooks
    # ext.exec_hooks(ext.OnPostIdentityInsert, identity)

    return identity


def update_identity(identity_id, identity_edit, old_identity):
    """Update the identity in the database.

    Update requests should be validated before being passed to this function.
    """
    t = tables.identity

    if 'custom_json' in identity_edit:
        identity_edit['custom_json'] = dict(
            old_identity['custom_json'],
            **identity_edit['custom_json']
        )

    # Update the "old" identity in place.
    stmt = (
        t.update()
        .where(t.c.id == identity_id)
        .values(**identity_edit)
        .returning(*list(t.c))
    )

    # Run on_pre_update hooks
    # ext.exec_hooks(ext.OnPreIdentityUpdate, old_identity, identity_edit)

    result = db.engine.execute(stmt).first()
    if not result:
        raise IdentityNotFoundError(
            'Identity {0} not found'.format(identity_id))

    identity = dict(result.items())

    # Run on_post_update hooks
    # ext.exec_hooks(ext.OnPostIdentityUpdate, old_identity, identity)

    return identity


def fetch_identity(identity_id):
    """Fetch an identity object by id from the database."""
    t = tables.identity
    stmt = t.select().where(t.c.id == identity_id)

    # TODO: Run on_pre_identity_fetch hooks
    # stmt = ext.exec_filter_hooks(ext.OnPreIdentityFetch, stmt)

    result = db.engine.execute(stmt).first()
    if not result:
        raise IdentityNotFoundError(
            'Identity {0} not found'.format(identity_id)
        )
    identity = dict(result.items())

    return identity


def update_comment(comment_id, comment_edit, old_comment,
                   update_modified=False):
    """Update the comment in the database in response to a request
    from the author. The `modified` timestamp will be set to the current time.

    Update requests should be validated before being passed to this function.
    """
    t = tables.comment

    if 'custom_json' in comment_edit:
        comment_edit['custom_json'] = dict(
            old_comment['custom_json'],
            **comment_edit['custom_json']
        )

    # Update the "old" comment in place.
    stmt = (
        t.update()
        .where(t.c.id == comment_id)
        .values(**comment_edit)
        .returning(*list(t.c))
    )

    # Update the `modified` timestamp if specified.
    if update_modified:
        stmt = stmt.values(modified=sa.text('NOW()'))

    # Run on_pre_update hooks
    ext.exec_hooks(ext.OnPreCommentUpdate, old_comment, comment_edit)

    result = db.engine.execute(stmt).first()
    if not result:
        raise CommentNotFoundError('Comment {0} not found'.format(comment_id))

    comment = dict(result.items())

    # Run on_post_update hooks
    ext.exec_hooks(ext.OnPostCommentUpdate, old_comment, comment)

    return comment


def validate_parent_exists(parent):
    """Validate that the parent exists in the database."""
    t = tables.comment
    stmt = sa.select([sa.exists([1]).where(t.c.id == parent)])
    return db.engine.execute(stmt).scalar()


def insert_identity_comment(identity_comment):
    """Insert the a new identity-to-comment object in to the database."""
    t = tables.identity_comment
    stmt = (
        t.insert()
        .values(**identity_comment)
        .returning(*list(t.c))
    )
    result = db.engine.execute(stmt).first()
    identity_comment = dict(result.items())

    # TODO: Run on_post_identity_comment_insert hooks
    # ext.exec_hooks(ext.OnPostIdentityInsert, identity)

    return identity_comment


def cte_chain(statements):
    """Chain a sequence of statements using a CTE. The result of the last
    statement will be returned when RETURNING is used."""
    parts = []
    bindparams = {}
    for i, stmt in enumerate(statements):
        compiled = stmt.compile(
            dialect=sqlalchemy.dialects.postgresql.dialect())
        # If first statement, open the WITH statement and use an alias.
        if i == 0:
            part = "WITH t{0} AS ( {1} )".format(i, compiled)
        # If last statement, do not prefix with comma and do not use an alias.
        elif i == (len(statements) - 1):
            part = "{0}".format(compiled)
        # Otherwise, if not first or last, prefix with comma and use an alias.
        else:
            part = ", t{0} AS ( {1} )".format(i, compiled)

        bindparams = utils.merge_fail_on_conflict(bindparams, compiled.params)
        parts.append(part)

    stmt = '\n'.join(parts)
    return stmt, bindparams



import flask_cors

from pg_discuss import ext


class CorsExt(ext.AppExtBase):
    """Enable Cross-Origin Resource Sharing via Flask-CORS."""
    def init_app(self, app):
        flask_cors.CORS(app, supports_credentials=True)



"""Minimal comment CRUD views.

 - fetch: fetch the list of comments for a thread
 - new: create a new comment
 - view: fetch a single comment
 - edit: update an existing comment
 - delete: delete a comment
"""
import flask
import flask_login
import functools

from . import queries
from . import forms
from . import serialize
from . import ext
from . import auth_forms


def check_mimetype(f):
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        if flask.request.mimetype != 'application/json':
            reason = (
                'Mimetype checking failed - Content-Type not set '
                'to application/json'
            )
            return flask.abort(400, reason)
        else:
            return f(*args, **kwargs)
    return wrapper


def fetch(thread_cid):
    """View to fetch the thread and it's comment collection as JSON."""
    raw_thread = queries.fetch_thread_by_client_id(thread_cid)
    # If the thread has not yet been created, return an empty JSON object.
    if not raw_thread:
        return flask.jsonify({})
    comments_seq = queries.fetch_comments_by_thread_client_id(thread_cid)
    app = flask.current_app
    hook_map = app.hook_map
    renderer = app.comment_renderer
    comments_seq = [serialize._to_client_comment(
        hook_map, renderer, c) for c in comments_seq]
    client_thread = serialize.to_client_thread(raw_thread, comments_seq)
    return flask.jsonify(client_thread)


@check_mimetype
def new(thread_cid):
    """View to create a new thread."""
    # Extract whitelisted attributes from request JSON
    # Note that extensions have the opportunity to process/persist additional
    # attributes from the request JSON, since they have access to the
    # request and `new_comment` object
    json = flask.request.get_json()
    allowed_keys = ['parent_id', 'text']
    new_comment = {k: json[k] for k in allowed_keys if k in json}

    # Create empty `custom_json`, for extensions to populate.
    new_comment['custom_json'] = {}

    # Associate comment with identity
    if hasattr(flask.g, 'identity'):
        new_comment['identity_id'] = flask.g.identity['id']
    # Set identity to None if no identity exists in request
    else:
        new_comment['identity_id'] = None

    # Validate required, type, text length
    new_comment = forms.validate_new_comment(new_comment)

    # Try to fetch the existing thread record, and create a new thread
    # if one is not found.
    thread = queries.fetch_thread_by_client_id(thread_cid)
    if not thread:
        thread = queries.insert_thread({'client_id': thread_cid})

    new_comment['thread_id'] = thread['id']

    # Insert the comment
    raw_comment = queries.insert_comment(new_comment)
    client_comment = serialize.to_client_comment(raw_comment)

    resp = flask.jsonify(client_comment)
    resp.status_code = 201

    ext.exec_hooks(ext.OnNewCommentResponse, resp, raw_comment, client_comment)

    return resp


def view(comment_id):
    """View to fetch a single comment."""
    # Fetch the comment
    raw_comment = queries.fetch_comment_by_id(comment_id)

    # Check if client has request plain (not marked-up) text
    plain = flask.request.args.get('plain') == '1'

    comment = serialize.to_client_comment(raw_comment, plain)
    return flask.jsonify(comment)


@check_mimetype
def edit(comment_id):
    """View to edit an existing comment."""
    json = flask.request.get_json()
    comment_edit = {'text': json['text']}

    # Create empty `custom_json`, for extensions to populate.
    # This will later be merged in to the existing `custom_json`.
    comment_edit['custom_json'] = {}

    # Validate required, type, text length
    # Use the id given in the URL path, ignoring any in the request JSON
    comment_edit = forms.validate_comment_edit(comment_edit)

    # Fetch the "old" comment
    old_comment = queries.fetch_comment_by_id(comment_id)

    # Check if does not belong to requesting identity
    if (
        not hasattr(flask.g, 'identity')
        or old_comment['identity_id'] != flask.g.identity['id']
    ):
        flask.abort(400,
                    'Cannot edit comment: comment belongs to another identity')

    # Check if deleted
    if old_comment['custom_json'].get('deleted'):
        flask.abort(400, 'Cannot edit comment: comment has been deleted')

    # Update the comment
    raw_comment = queries.update_comment(
        comment_id,
        comment_edit,
        old_comment,
        update_modified=True
    )
    comment = serialize.to_client_comment(raw_comment)
    return flask.jsonify(comment)


@check_mimetype
def delete(comment_id):
    """Mark a comment as deleted. The comment will still show up in API results,
    but only containing values for it's `id`, `thread_id`, and `parent_id`.

    Note that the actual database record is not deleted. Deletes are handled
    in the same way that edits are: the record is updated in place, and,
    if using the `archive_updated` extension, an archive is made of the old
    record. The `deleted` flag may be used to indicate special rendering of
    deleted comments.
    """
    comment_edit = {
        'text': '',
        'custom_json': {
            'deleted': True,
        }
    }

    # Fetch the "old" comment
    old_comment = queries.fetch_comment_by_id(comment_id)

    # Check if does not belong to requesting identity
    if (
        not hasattr(flask.g, 'identity')
        or old_comment['identity_id'] != flask.g.identity['id']
    ):
        flask.abort(
            400,
            'Cannot delete comment: this comment belongs to another identity')

    # Check if already deleted
    if old_comment['custom_json'].get('deleted'):
        flask.abort(
            400, 'Cannot delete comment: comment has already been deleted')

    # Mark the comment as deleted
    result = queries.update_comment(
        comment_id,
        comment_edit,
        old_comment,
        update_modified=True
    )
    return flask.jsonify(result)


def landing():
    return flask.render_template('landing.html')


def admin_login():
    """View to handle Admin logins."""
    form = auth_forms.LoginForm(flask.request.form)
    if form.validate_on_submit():
        # Login and validate the user.
        user = form.get_user()
        flask_login.login_user(user)
        return flask.current_app.login_callback()
    return flask.render_template('login.html', form=form)


def admin_logout():
    """View to handle Admin logouts."""
    flask_login.logout_user()
    return 'Logged out successfully'



"""Functions to prepare comments and threads for serialization."""
import flask
import markupsafe

from . import ext
from . import _compat

# Fields which will be serialized by default when sent to client.
DEFAULT_COMMENT_WHITELIST = [
    'id',
    'thread_id',
    'parent_id',
    'created',
    'modified',
    'text',
]

DEFAULT_THREAD_WHITELIST = [
    'id',
    'client_id',
]


def to_client_comment(raw_comment, plain=False):
    """A convenience wrapper for `_to_client_comment`. Should be used
    when only single comments are processed within a request.

    For bulk comment processing, there is a significant speedup to using
    _to_client_comment directly. See :func:`pg_discuss.views.fetch` for
    an example.
    """
    app = flask.current_app
    hook_map = app.hook_map
    renderer = app.comment_renderer
    return _to_client_comment(hook_map, renderer, raw_comment, plain)


def _to_client_comment(hook_map, renderer, raw_comment, plain=False):
    """Prepare comments for serialization to JSON.

    Only preserves whitelisted attributes. Calls any `OnPreCommentSerialize`
    extensions, and the CommentRenderer driver.

    Takes `app` as an arg to eliminate repeated `current_app` lookups.

    This is a performance critical function - it is called for every
    comment when a thread is fetched.
    """
    client_comment = {k: raw_comment[k] for k in DEFAULT_COMMENT_WHITELIST}

    # Extract `deleted` attribute from custom json as well.
    if 'deleted' in raw_comment['custom_json']:
        client_comment['deleted'] = raw_comment['custom_json']['deleted']

    # Run on_comment_serialize hooks. Inline the `exec_hooks` function
    # for performance.
    args = (raw_comment, client_comment)
    for ext_obj in hook_map[ext.OnPreCommentSerialize]:
        ext_obj.on_pre_comment_serialize(*args)

    # Escape string fields, besides `text`, which may be rendered into DOM.
    # We have no way of knowing what fields extensions might add, so we need
    # to do isinstance checks to see which are strings.
    for k, v in client_comment.items():
        if isinstance(v, _compat.string_types) and k != 'text':
            client_comment[k] = markupsafe.escape(v)

    # Render comment text using configured CommentRenderer. Th renderer should
    # handle escaping of the comment text.
    if not plain:
        client_comment['text'] = renderer.render(client_comment['text'])

    return client_comment


def to_client_thread(raw_thread, comment_seq):
    """Prepare thread and it's comment collection for serialization to JSON.

    Only preserves whitelisted attributes. Calls any `OnPreThreadSerialize`
    extensions.

    Return value is an object whose JSON representation is the response.
    """
    client_thread = {k: raw_thread[k] for k in DEFAULT_THREAD_WHITELIST}

    client_thread['comments'] = comment_seq

    # Run on_pre_thread_serialize hooks
    ext.exec_hooks(ext.OnPreThreadSerialize, raw_thread, comment_seq,
                   client_thread)

    return client_thread



# flake8: noqa
from setuptools import setup, find_packages

requires = [
    'pytz>=2015.6',
    'Flask-Admin>=1.3.0',
    'Flask-Mail>=0.9.1',
]

if __name__ == '__main__':
    setup(
        name='pg-discuss-blessed-extensions',
        version='1.0',

        description='Blessed extensions for pg-discuss',
        license='MIT',

        author='Steffen Prince',
        author_email='steffen@sprin.io',

        url='https://sprin.io',
        download_url='https://github.com/sprin/pg-discuss',

        classifiers=['Development Status :: 3 - Alpha',
                     'License :: OSI Approved :: MIT License',
                     'Programming Language :: Python',
                     'Programming Language :: Python :: 2',
                     'Programming Language :: Python :: 2.7',
                     'Programming Language :: Python :: 3',
                     'Programming Language :: Python :: 3.4',
                     'Intended Audience :: Developers',
                     'Environment :: Web Environment',
                     'Framework :: Flask',
                     ],

        platforms=['Any'],

        scripts=[],

        provides=['pg_discuss.ext',
                  ],

        packages=find_packages(),
        include_package_data=True,

        entry_points={
            'pg_discuss.ext': [
                'blessed_csrf_token = blessed_extensions.csrf_token:CsrfTokenExt',
                'blessed_route_list = blessed_extensions.route_list:RouteListExt',
                'blessed_archive_comment_versions = blessed_extensions.archive_comment_versions:ArchiveCommentVersionsExt',
                'blessed_validate_comment_len = blessed_extensions.validate_comment_len:ValidateCommentLen',
                'blessed_unix_time_json_encoder = blessed_extensions.unix_time_json_encoder:UnixTimeJSONEncoder',
                'blessed_isso_client_shim = blessed_extensions.isso_client_shim:IssoClientShim',
                'blessed_cors = blessed_extensions.cors:CorsExt',
                'blessed_capture_author = blessed_extensions.capture_author:CaptureAuthor',
                'blessed_capture_website = blessed_extensions.capture_website:CaptureWebsite',
                'blessed_capture_email = blessed_extensions.capture_email:CaptureEmail',
                'blessed_capture_remote_addr = blessed_extensions.capture_remote_addr:CaptureRemoteAddr',
                'blessed_auth_tkt_identity_policy = blessed_extensions.auth_tkt_identity_policy:AuthTktIdentityPolicy',
                'blessed_persist_comment_info_on_id = blessed_extensions.auth_tkt_identity_policy:PersistCommentInfoOnIdentity',
                'blessed_markdown_renderer = blessed_extensions.markdown_renderer:MarkdownRenderer',
                'blessed_voting = blessed_extensions.voting:Voting',
                'blessed_admin = blessed_extensions.admin:AdminExt',
                'blessed_moderation = blessed_extensions.moderation:ModerationExt',
                'blessed_mod_email = blessed_extensions.mod_email:ModerationEmail',
                'blessed_profiler = blessed_extensions.profiler:ProfilerExt',
                'blessed_proxyfix = blessed_extensions.proxyfix:ProxyFixExt',
            ],
        },

        zip_safe=False,

        install_requires=requires,
    )



"""Factory function for the WSGI application object.
"""
import os
import sys

import flask
import flask_login
import flask_migrate
import flask_script
import stevedore

from . import _compat
from . import auth_forms
from . import config
from . import ext
from . import identity
from . import models
from . import views
from .db import db

if _compat.PYPY:  # pragma: no cover
    from psycopg2cffi import compat as pg2cfficompat


def app_factory():
    """Factory function for the WSGI application object.

    Reads config, initializes Flask extensions that are part of the core,
    loads pg-config drivers and extensions, and configures core views.
    """
    # Use psycopg2cffi if PYPY
    if _compat.PYPY:  # pragma: no cover
        pg2cfficompat.register()

    # Flask application
    app = flask.Flask('pg-discuss')
    app.root_path = os.path.abspath(os.path.dirname(__file__))

    # Load default config values from pg_discuss.config module
    app.config.from_object(config)

    # Load custom config from user-defined PG_DISCUSS_SETTINGS_FILE
    custom_settings = os.environ.get('PG_DISCUSS_SETTINGS_FILE',
                                     '/opt/pg-discuss/local_settings.py')
    if custom_settings and os.path.isfile(custom_settings):
        app.config.from_pyfile(custom_settings)

    # Set the recursion limit
    sys.setrecursionlimit(app.config['PYTHON_RECURSION_LIMIT'])

    # Flask-SQLAlchemy
    db.init_app(app)

    # Flask-Migrate
    app.migrate = flask_migrate.Migrate(app, db)

    # Flask-Script
    app.script_manager = flask_script.Manager(app)
    app.script_manager.add_command('db', flask_migrate.MigrateCommand)
    app.script_manager.add_command('createadminuser',
                                   auth_forms.CreateAdminUser)

    # Flask-Login, for Admin users.
    app.admin_login_manager = flask_login.LoginManager(app)
    # Set up callback to load user objects`
    app.admin_login_manager.user_loader(
        lambda user_id: db.session.query(models.AdminUser).get(user_id))

    # Use stevedore to load drivers/extensions.
    # Discover all drivers/extensions, but do not load any.
    # Used for logging found extensions.
    app.ext_mgr_all = stevedore.ExtensionManager(namespace='pg_discuss.ext')

    # Load configured IdentityPolicy driver
    app.identity_policy_loader = stevedore.DriverManager(
        namespace='pg_discuss.ext',
        name=app.config['DRIVER_IDENTITY_POLICY'],
    )
    # Initialize IdentityPolicyManager with configured IdentityManager
    app.identity_policy_mgr = identity.IdentityPolicyManager(
        app,
        app.identity_policy_loader.driver,
    )

    # Load configured CommentRenderer driver
    app.comment_renderer_loader = stevedore.DriverManager(
        namespace='pg_discuss.ext',
        name=app.config['DRIVER_COMMENT_RENDERER'],
    )
    app.comment_renderer = app.comment_renderer_loader.driver(app)

    # Load configured JSONEncoder driver
    app.json_encoder_loader = stevedore.DriverManager(
        namespace='pg_discuss.ext',
        name=app.config['DRIVER_JSON_ENCODER'],
    )
    app.json_encoder = app.json_encoder_loader.driver

    # Exempt public read-only views from IdentityPolicy
    app.identity_policy_mgr.exempt(views.fetch)
    app.identity_policy_mgr.exempt(views.view)

    # Default routes. Other routes must be added through App extensions.
    # Default routes are set up before app extensions are loaded so extensions
    # can introspect/modify view functions.
    app.route('/threads/<thread_cid>/comments', methods=['GET'])(views.fetch)
    app.route('/threads/<thread_cid>/comments', methods=['POST'])(views.new)
    app.route('/comments/<int:comment_id>', methods=['GET'])(views.view)
    app.route('/comments/<int:comment_id>', methods=['PATCH'])(views.edit)
    app.route('/comments/<int:comment_id>', methods=['DELETE'])(views.delete)
    app.route('/login', methods=['GET', 'POST'])(views.admin_login)
    app.route('/logout', methods=['GET'])(views.admin_logout)

    # Load all extensions explicitly enabled via `ENABLE_EXT_*` parameters.
    app.ext_mgr = stevedore.NamedExtensionManager(
        namespace='pg_discuss.ext',
        names=config.get_enabled_extensions(app.config),
        name_order=True,
        invoke_on_load=True,
        invoke_kwds={'app': app},
        on_load_failure_callback=ext.fail_on_ext_load,
        propagate_map_exceptions=True,
    )
    # Create hook map
    app.hook_map = ext.get_hook_map(app.ext_mgr.extensions, ext.hook_classes())

    # Run the `init_app` hooks.
    ext.exec_init_app(app)

    # Add a route to the landing page at the root, '/'. Ignore if an extension
    # has already set up a route for the root.
    try:
        app.route('/', methods=['GET'])(views.landing)
    except:
        pass

    return app



import hashlib
import hmac
import os
import time

import flask
import werkzeug.security

from pg_discuss import _compat
from pg_discuss import ext

#: Protect all views from CSRF by verifying token by default.
CSRF_TOKEN_CHECK_DEFAULT = True
#: Expiration of a CSRF token in seconds.
CSRF_TOKEN_TIME_LIMIT = 3600
#: Headers in which to submit CSRF token.
CSRF_TOKEN_HEADERS = ['X-CSRF-Token']
#: Exempt a list of HTTP methods from the check (read-only methods should be
#: exempted).
CSRF_TOKEN_EXEMPT_METHODS = ['GET', 'HEAD', 'OPTIONS', 'TRACE']
CSRF_SSL_STRICT = True


class CsrfTokenExt(ext.AppExtBase):
    """
    CSRF token extension.

    Forked from flask_wtf.csrf:
    https://github.com/lepture/flask-wtf/blob/HEAD/flask_wtf/csrf.py

    Verifies the token present in the cookie against a token sent in the
    `X-CSRF-Token` HTTP header.
    """
    def init_app(self, app):
        self._app = app
        self._exempt_views = []
        app.config.setdefault('CSRF_TOKEN_CHECK_DEFAULT',
                              CSRF_TOKEN_CHECK_DEFAULT)
        app.config.setdefault('CSRF_TOKEN_TIME_LIMIT', CSRF_TOKEN_TIME_LIMIT)
        app.config.setdefault('CSRF_TOKEN_HEADERS', CSRF_TOKEN_HEADERS)
        app.config.setdefault('CSRF_TOKEN_EXEMPT_METHODS',
                              CSRF_TOKEN_EXEMPT_METHODS)
        app.config.setdefault('CSRF_SSL_STRICT', CSRF_SSL_STRICT)

        if not app.config['CSRF_TOKEN_CHECK_DEFAULT']:
            return

        @app.after_request
        def add_token(response):
            """Add the token to *every* response. We don't necessarily know
            which endpoints the client will access and expect to get a
            token in the response, so it is simplest if we always return a
            token.
            """
            token = generate_csrf()
            for header_name in CSRF_TOKEN_HEADERS:
                response.headers[header_name] = token
            return response

        @app.before_request
        def _csrf_protect():
            # many things come from django.middleware.csrf
            if flask.request.method in app.config['CSRF_TOKEN_EXEMPT_METHODS']:
                return

            if self._exempt_views:
                if not flask.request.endpoint:
                    return

                view = app.view_functions.get(flask.request.endpoint)
                if not view:
                    return

                dest = '%s.%s' % (view.__module__, view.__name__)
                if dest in self._exempt_views:
                    return

            self.protect()

    def _get_csrf_token(self):
        """Extract CSRF token from headers.
        """
        for header_name in self._app.config['CSRF_TOKEN_HEADERS']:
            csrf_token = flask.request.headers.get(header_name)
            if csrf_token:
                return csrf_token
        return None

    def protect(self):
        if (
            flask.request.method
            in self._app.config['CSRF_TOKEN_EXEMPT_METHODS']
        ):
            return

        if not validate_csrf(self._get_csrf_token()):
            reason = 'CSRF token missing or incorrect.'
            return self._error_response(reason)

        if flask.request.is_secure and self._app.config['CSRF_SSL_STRICT']:
            if not flask.request.referrer:
                reason = 'Referrer checking failed - no Referrer.'
                return self._error_response(reason)

            good_referrer = 'https://%s/' % flask.request.host
            if not same_origin(flask.request.referrer, good_referrer):
                reason = 'Referrer checking failed - origin does not match.'
                return self._error_response(reason)

        # Mark the request as having a valid CSRF token
        flask.request.csrf_token_valid = True

    def exempt(self, view):
        """Exclude a view from csrf protection. Takes a view function as the
        single argument.
        """
        view_location = '%s.%s' % (view.__module__, view.__name__)
        self._exempt_views.append(view_location)

    def _error_response(self, reason):
        return flask.abort(403, reason)


def generate_csrf(secret_key=None, time_limit=None):
    """Generate csrf token code.
    :param secret_key: A secret key for mixing in the token,
                       default is Flask.secret_key.
    :param time_limit: Token valid in the time limit,
                       default is 3600s.
    """
    if not secret_key:
        secret_key = flask.current_app.config['SECRET_KEY']
        if not secret_key:
            raise ValueError('Must provide secret_key to use csrf.')

    if time_limit is None:
        time_limit = flask.current_app.config['CSRF_TOKEN_TIME_LIMIT']

    if 'csrf_token' not in flask.session:
        flask.session['csrf_token'] = hashlib.sha1(os.urandom(64)).hexdigest()

    if time_limit:
        expires = int(time.time() + time_limit)
        csrf_build = '%s%s' % (flask.session['csrf_token'], expires)
    else:
        expires = ''
        csrf_build = flask.session['csrf_token']

    hmac_csrf = hmac.new(
        _compat.to_bytes(secret_key),
        _compat.to_bytes(csrf_build),
        digestmod=hashlib.sha1
    ).hexdigest()
    return '%s##%s' % (expires, hmac_csrf)


def validate_csrf(data, secret_key=None, time_limit=None):
    """Check if the given data is a valid csrf token.
    :param data: The csrf token value to be checked.
    :param secret_key: A secret key for mixing in the token,
                       default is Flask.secret_key.
    :param time_limit: Check if the csrf token is expired.
                       default is True.
    """
    if not data or '##' not in data:
        return False

    expires, hmac_csrf = data.split('##', 1)

    if time_limit is None:
        time_limit = flask.current_app.config['CSRF_TOKEN_TIME_LIMIT']

    if time_limit:
        try:
            expires = int(expires)
        except ValueError:
            return False

        now = int(time.time())
        if now > expires:
            return False

    if not secret_key:
        secret_key = flask.current_app.config['SECRET_KEY']
        if not secret_key:
            raise ValueError('Must provide secret_key to use csrf.')

    if 'csrf_token' not in flask.session:
        return False

    csrf_build = '%s%s' % (flask.session['csrf_token'], expires)
    hmac_compare = hmac.new(
        _compat.to_bytes(secret_key),
        _compat.to_bytes(csrf_build),
        digestmod=hashlib.sha1
    ).hexdigest()

    return werkzeug.security.safe_str_cmp(hmac_compare, hmac_csrf)


def same_origin(current_uri, compare_uri):
    parsed_uri = _compat.urlparse(current_uri)
    parsed_compare = _compat.urlparse(compare_uri)

    if parsed_uri.scheme != parsed_compare.scheme:
        return False

    if parsed_uri.hostname != parsed_compare.hostname:
        return False

    if parsed_uri.port != parsed_compare.port:
        return False
    return True



import os
from threading import Thread

import flask
import flask_mail
import sqlalchemy as sa

from pg_discuss import tables
from pg_discuss import ext
from pg_discuss.db import db

#: Host to use for sending email.
MAIL_HOST = None
#: Port to use for sending email.
MAIL_PORT = None
#: Username to use to authenticate with the SMTP server.
MAIL_USERNAME = None
#: Password to use to authenticate with the SMTP server.
MAIL_PASSWORD = None
#: Use an explicit secure connection using TLS/SSL negotiation.
#: Typically uses port 465.
#:
#: Internally, this setting is translated to Flask-Mail's `MAIL_USE_SSL`
#: setting.
MAIL_SECURE_CONNECTION = True
#: Use STARTTLS. This is a broken protocol, trivially exploitable by MitM.
#: Use an explicit TLS connection instead if possible.
#: See: https://www.agwa.name/blog/post/starttls_considered_harmful
#:
#: Typically uses port 587.
#:
#: `MAIL_SECURE_CONNECTION` and `MAIL_USE_STARTTLS` are mutually exclusive,
#: and an error will be raised if both are set to True.
#:
#: Internally, this setting is translated to Flask-Mail's `MAIL_USE_TLS`
#: setting.
MAIL_USE_STARTTLS = False
#: Send mail asynchronously by spawning a thread. Dramatically improves
#: response times for posting new comments.
MAIL_SEND_ASYNC = True


class ModerationEmail(ext.AppExtBase, ext.OnPostCommentInsert):
    """Extensions to send a mail to the administrators when new comments are
    submitted. All registered administrators will receive a mail.

    Include a link to the Admin panel.
    """

    mail_settings_from_env = [
        'MAIL_HOST',
        'MAIL_PORT',
        'MAIL_USERNAME',
        'MAIL_PASSWORD',
        'MAIL_SECURE_CONNECTION',
        'MAIL_USE_STARTTLS',
    ]

    def init_app(self, app):
        app.config.setdefault('MAIL_HOST', MAIL_HOST)
        app.config.setdefault('MAIL_PORT', MAIL_PORT)
        app.config.setdefault('MAIL_USERNAME', MAIL_USERNAME)
        app.config.setdefault('MAIL_PASSWORD', MAIL_PASSWORD)
        app.config.setdefault('MAIL_SECURE_CONNECTION', MAIL_SECURE_CONNECTION)
        app.config.setdefault('MAIL_USE_STARTTLS', MAIL_USE_STARTTLS)
        app.config.setdefault('MAIL_SEND_ASYNC', MAIL_SEND_ASYNC)

        # Get email connection settings from environment, by default.
        self.get_settings_from_env(app)
        if (
            app.config['MAIL_SECURE_CONNECTION']
            and app.config['MAIL_USE_STARTTLS']
        ):
            raise Exception(
                'Only one of MAIL_SECURE_CONNECTION and MAIL_USE_STARTTLS '
                'may be True')
        # Translate MAIL_SECURE_CONNECTION to Flask-Mail's MAIL_USE_SSL
        app.config['MAIL_USE_SSL'] = app.config['MAIL_SECURE_CONNECTION']
        # Translate MAIL_USE_STARTTLS to Flask-Mail's MAIL_USE_TLS
        app.config['MAIL_USE_TLS'] = app.config['MAIL_USE_STARTTLS']

        self.mail = flask_mail.Mail(app)

    def get_settings_from_env(self, app):
        for setting in self.mail_settings_from_env:
            val = os.environ.get(setting)
            if val:
                if setting in ['MAIL_SECURE_CONNECTION', 'MAIL_USE_STARTTLS']:
                    app.config[setting] = (val == 'True')
                else:
                    app.config[setting] = val

    def on_post_comment_insert(self, comment):
        subject = "pg-discuss: New comment from {}".format(
            comment['custom_json']['author'])
        body = (
            "Moderation panel:\n{url}\n\n{author} posted:\n\n{comment}"
            .format(
                url=flask.url_for('moderation.index_view', _external=True),
                author=comment['custom_json']['author'],
                comment=comment['text']
            ))
        msg = flask_mail.Message(subject=subject, body=body)
        # Send asynchronously in a thread, if enabled.
        if self.app.config['MAIL_SEND_ASYNC']:
            thr = Thread(target=send_async_email,
                         args=[self.app, msg, self.mail])
            thr.start()
        else:
            recipients = fetch_admin_emails()
            for recipient in recipients:
                msg.add_recipient(recipient)
            self.mail.send(msg)


def fetch_admin_emails():
    stmt = sa.select([tables.admin_user.c.email])
    result = db.engine.execute(stmt).fetchall()
    if result:
        return [x[0] for x in list(result)]


def send_async_email(app, msg, mail):
    with app.app_context():
        recipients = fetch_admin_emails(app)
        for recipient in recipients:
            msg.add_recipient(recipient)
        mail.send(msg)



import flask
import flask_admin
import flask_admin.contrib.sqla
import flask_admin.contrib.sqla.form
import flask_login
import os
import jinja2
import simplejson as json
import sqlalchemy.orm
import wtforms.fields
import werkzeug.utils

from pg_discuss import ext
from pg_discuss import models

THIS_DIR = os.path.dirname(os.path.abspath(__file__))

class AdminExt(ext.AppExtBase):
    """Extension to add an admin interface. Allows management of Comment,
    Thread, Identity, and AdminUser models.

    Available to authenticated Admin users.
    """
    def init_app(self, app):
        # Add template loader
        my_loader = jinja2.ChoiceLoader([
            app.jinja_loader,
            jinja2.FileSystemLoader(THIS_DIR),
        ])
        app.jinja_loader = my_loader

        # Add static file route. The web server should be configured to serve
        # static files, not the app, but we do it here to allow the files
        # to be served without additional web server config.
        app.route('/admin_static/<path:filename>', methods=['GET'])(
            self.admin_static)

        # Configure Admin
        admin = flask_admin.Admin(app, 'pg-discuss Administration',
                                  template_mode='bootstrap3',
                                  base_template='admin_templates/base.html',
                                  index_view=MyAdminIndexView())
        app.admin = admin
        # Add views
        admin.add_view(CommentAdmin(PrettyComment, models.db.session,
                                    name='Comment'))
        admin.add_view(ThreadAdmin(models.Thread, models.db.session))
        admin.add_view(IdentityAdmin(PrettyIdentity, models.db.session,
                                     name='Identity'))
        admin.add_view(AdminUserAdmin(models.AdminUser, models.db.session))

        # Set up login callback.
        app.login_callback = (
            lambda: flask.redirect(flask.url_for('admin.index')))

    def admin_static(self, filename):
        return flask.send_from_directory(
            THIS_DIR + '/admin_static/',
            werkzeug.utils.secure_filename(filename))


class PrettyIdentity(models.Identity):
    """Subclass the Identity model to provide a nicer string represention."""
    def __unicode__(self):
        names = self.custom_json.get('names')
        remote_addrs = self.custom_json.get('remote_addrs')
        if names:
            str_ = names[0]
        elif remote_addrs:
            str_ = remote_addrs[0]
        else:
            str_ = '?'
        return "({}) {}".format(self.id, str_)
    __str__ = __unicode__


class PrettyComment(models.Comment):
    """Subclass the Identity model to provide a nicer string represention."""
    identity_ = sqlalchemy.orm.relationship('PrettyIdentity')


class AuthenticatedModelView(flask_admin.contrib.sqla.ModelView):
    def is_accessible(self):
        return flask_login.current_user.is_authenticated

    def inaccessible_callback(self, name, **kwargs):
        # redirect to login page if user doesn't have access
        return flask.redirect(flask.url_for('admin_login'))


class DictToJSONField(wtforms.fields.TextAreaField):
    def process_data(self, value):
        if value is None:
            value = {}

        self.data = json.dumps(value, sort_keys=True,
                               indent=4, separators=(',', ': '))

    def process_formdata(self, valuelist):
        if valuelist:
            self.data = json.loads(valuelist[0])
        else:
            self.data = '{}'


class CustomAdminConverter(flask_admin.contrib.sqla.form.AdminModelConverter):
    @flask_admin.contrib.sqla.form.converts('JSON')
    def conv_JSON(self, field_args, **extra):
        return DictToJSONField(**field_args)


class CommentAdmin(AuthenticatedModelView):
    model_form_converter = CustomAdminConverter
    column_exclude_list = ['identity']
    form_excluded_columns = ['identity']
    form_widget_args = {
        'custom_json': {'rows': 10}
    }


class ThreadAdmin(AuthenticatedModelView):
    model_form_converter = CustomAdminConverter
    form_widget_args = {
        'custom_json': {'rows': 10}
    }


class IdentityAdmin(AuthenticatedModelView):
    model_form_converter = CustomAdminConverter
    form_widget_args = {
        'custom_json': {'rows': 10}
    }


class AdminUserAdmin(AuthenticatedModelView):
    pass


class MyAdminIndexView(flask_admin.AdminIndexView):

    @flask_admin.expose('/')
    def index(self):
        if not flask_login.current_user.is_authenticated:
            return flask.redirect(flask.url_for('admin_login'))
        return super(MyAdminIndexView, self).index()



"""The `blessed_extensions` package contains extensions and drivers which are
bundled with pg-discuss.

Installation is optional, but recommended since the
functionality contained in the core is probably too minimal for most
installations. However, some deployments may want to supply all of their own
drivers and extensions, and in this case `blessed_extensions` does not need
to be installed.

pg-discuss uses uses setuptools entrypoints for discovery/loading of extensions
and drivers. The entrypoints are defined in `setup.py`.
"""



import flask

from pg_discuss import ext
from pg_discuss import queries


class AuthTktIdentityPolicy(ext.IdentityPolicy):
    """An IdentityPolicy tied to a browser session via a cookie.

    This policy does not provide for any authentication, but simply
    associates an identity with a browser session so that a user may
    edit/delete comments associated with their browser session.

    This policy does not attempt to enforce that users supply a consistent
    name/email for each comment. It does persist names and emails with
    the Identity record for the sake of labeling identities, but otherwise,
    names/emails have no functional importance with this policy.
    """

    def __init__(self, app):
        self.app = app
        # Enable the related extension to persist comment info
        # with the session identity.
        app.config.setdefault('ENABLE_EXT_BLESSED_PERSIST_COMMENT_INFO_ON_ID',
                              True)

    def remember(self, request, identity_id, **extras):
        # Update cookie
        flask.session['identity_id'] = identity_id

    def get_identity(self, request, **extras):
        identity_id = flask.session.get('identity_id')
        from flask import current_app
        if identity_id:
            try:
                return queries.fetch_identity(identity_id)
            # If we have a cookie, but no matching identity in the db,
            # create a new identity and reset cookie.
            except queries.IdentityNotFoundError:
                return queries.insert_identity()
        else:
            current_app.logger.error(identity_id)
            return queries.insert_identity()

    def forget(self, request, **extras):
        flask.session.pop('identity_id')


class PersistCommentInfoOnIdentity(ext.OnPostCommentInsert):

    def on_post_comment_insert(self, comment):
        """Update Identity record with name and email entered for comment.
        Since there is not necessarily an enforcement that names and emails are
        required or consistent across comments from the same identity, we treat
        names and emails as a set.

        If name/email are not captured, then this is a NOP.
        """
        new_name = comment['custom_json'].get('author')
        known_names = flask.g.identity['custom_json'].get('names', [])
        identity_edit = {'custom_json': {}}
        if new_name and new_name not in known_names:
            known_names.append(new_name)
            identity_edit['custom_json']['names'] = known_names

        new_remote_addr = comment['custom_json'].get('remote_addr')
        known_remote_addrs = flask.g.identity['custom_json'].get(
            'remote_addrs', [])
        if new_remote_addr and new_remote_addr not in known_remote_addrs:
            known_remote_addrs.append(new_remote_addr)
            identity_edit['custom_json']['remote_addrs'] = known_remote_addrs

        new_email = comment['custom_json'].get('email')
        known_emails = flask.g.identity['custom_json'].get('emails', [])
        if new_email and new_email not in known_emails:
            known_emails.append(new_email)
            identity_edit['custom_json']['emails'] = known_emails

        if identity_edit['custom_json']:
            flask.g.identity = queries.update_identity(
                flask.g.identity['id'],
                identity_edit,
                flask.g.identity,
            )



"""Core JSON encoder. Converts DateTime object to ISO 8601 date strings.

Note that subclassing `JSONEncoder` is not recommended by the author of
`simplejson` since it is possible to have conflicts with attributes on the
parent class, or depend on attributes on the parent class that may be changed.
Also, the author suggests subclassing will be faster since `JSONEncoder` can
perform "caching of instances when certain arguments are not overriden." The
author's recommendation to implement the strategy pattern for JSON encoders is
to implement the strategy as a "default" callable that is passed to
`JSONEncoder` as an argument.
See: https://github.com/simplejson/simplejson/issues/124

However, subclasses which provide a new `default` method without acessing or
modifying other `JSONEncoder` internals should not have these drawbacks.
Timing of this `JSONEncoder` subclass did not show any difference compared to
providing a "default" callable.

Finally, Flask offers some JSON helpers, and assumes the user
will set `app.json_encoder` to a `JSONEncoder` subclass if a custom encoder
is needed. Therefore, pg-discuss will likewise expect subclasses of
`JSONEncoder` for the value of `DRIVER_JSON_ENCODER`.
"""
import datetime

import simplejson as json


class IsoDateJSONEncoder(json.JSONEncoder):
    """Convert datetime objects to ISO 8601 strings. The exact format is
    "YYYY-MM-DDTHH:MM:SS.mmmmmmZ", which can be parsed directly by the
    JavaScript Date constructor.
    """
    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            return obj.isoformat() + 'Z'
        return json.JSONEncoder.default(self, obj)



import sys
import os
from setuptools import setup
from setuptools.command.develop import develop

PYPY = hasattr(sys, 'pypy_version_info')

requires = [
    'SQLAlchemy>=1.0',
    'flask>=0.10, <1.0',
    'alembic>=0.8.0',
    'Flask-SQLAlchemy>=2.0',
    'Flask-Script>=2.0',
    'Flask-Migrate>=1.5.0',
    'Flask-Login>=0.3.0',
    'Flask-WTF>=0.12',
    'voluptuous>=0.8',
    'stevedore>=1.7',
    'flask-cors>=2.1',
    'simplejson>=3.8',
    'misaka>=2.0.0',
]

if PYPY:
    requires += ['psycopg2cffi>=2.7']
else:
    requires += ['psycopg2>=2.6']


def get_data_files_list(paths):
    data_files = []
    for path in paths:
        for root, dirnames, filenames in os.walk(path):
            if not root.endswith('__pycache__'):
                full_filenames = [
                    os.path.join(root, f) for f in filenames
                    if not f.endswith('.pyc')]
                data_files.append([root, full_filenames])
    return data_files


class DevelopDocs(develop):
    """Custom command to require docs dependencies."""

    def __init__(self, *args, **kwargs):
        global requires
        requires += ['sphinx>=1.3.1']
        develop.__init__(self, *args, **kwargs)


class DevelopTests(develop):
    """Custom command to require test dependencies."""

    def __init__(self, *args, **kwargs):
        global requires
        requires += [
            'pytest>=2.8.0',
            'pytest-cov>=2.1.0',
        ]
        develop.__init__(self, *args, **kwargs)

setup(
    name='pg-discuss',
    version='1.0b1',

    description='A comment system backend on top of PostgreSQL',
    license='MIT',

    author='Steffen Prince',
    author_email='steffen@sprin.io',

    url='https://pg-discuss.sprin.io',
    download_url='https://github.com/sprin/pg-discuss',

    classifiers=['Development Status :: 4 - Beta',
                 'License :: OSI Approved :: MIT License',
                 'Programming Language :: Python',
                 'Programming Language :: Python :: 2',
                 'Programming Language :: Python :: 2.7',
                 'Programming Language :: Python :: 3',
                 'Programming Language :: Python :: 3.4',
                 'Programming Language :: Python :: 3.5',
                 'Intended Audience :: Developers',
                 'Environment :: Web Environment',
                 'Framework :: Flask',
                 ],

    platforms=['Any'],

    provides=['pg_discuss'],
    packages=['pg_discuss', 'pg_discuss.drivers'],
    include_package_data=True,
    package_data={'pg_discuss': ['*.html', '*.js', '*.svg']},
    data_files=(
        get_data_files_list(['isso', 'migrations', 'ext_migrations',
                             'blessed_extensions'])
        + [['', ['uwsgi.ini', 'main.py']]]),

    # Entrypoints for drivers included in core. We need to include basic
    # drivers in core so that the app is functional without the
    # `blessed_extensions` package or additional driver packages.
    entry_points={
        'pg_discuss.ext': [
            'core_escaping_renderer = pg_discuss.drivers.escaping_renderer:EscapingRenderer',
            'core_null_identity_policy = pg_discuss.drivers.null_identity_policy:NullIdentityPolicy',
            'core_iso_date_json_encoder = pg_discuss.drivers.iso_date_json_encoder:IsoDateJSONEncoder',
        ],
        'console_scripts': [
            'pgd-admin = pg_discuss.management:execute_from_command_line',
        ],
    },

    zip_safe=False,

    install_requires=requires,

    cmdclass={
        'develop_docs': DevelopDocs,
        'develop_tests': DevelopTests,
    },
)



from werkzeug.contrib.fixers import ProxyFix

from pg_discuss import ext

#: Number of HTTP proxies the server is running behind.
NUM_PROXIES = 1


class ProxyFixExt(ext.AppExtBase):
    """Enable the ProxyFix WSGI middleware.

    See:
    http://flask.pocoo.org/docs/0.10/deploying/wsgi-standalone/#proxy-setups

    Do not enable this middleware if the application is not running behind
    a proxy - this will render the app vulnerable to IP spoofing.
    """
    def __init__(self, app):
        self.app = app
        app.config.setdefault('NUM_PROXIES', NUM_PROXIES)

    def init_app(self, app):
        app.wsgi_app = ProxyFix(app.wsgi_app,
                                num_proxies=self.app.config['NUM_PROXIES'])



import flask
import flask_script

from pg_discuss import _compat
from pg_discuss import ext


class RouteListExt(ext.AppExtBase):
    """Extension to add a command to list the currently configured routes
    in the application.

    Useful since extensions can add routes at runtime.
    """

    def init_app(self, app):
        app.script_manager.add_command('list_routes', ListRoutes())


class ListRoutes(flask_script.Command):
    """List all routes configured on application.
    """

    def run(self):
        """List all routes configured on application.
        """
        output = []
        for rule in flask.current_app.url_map.iter_rules():
            methods = ','.join(rule.methods)
            line = _compat.unquote("{:50s} {:20s} {}".format(
                rule.endpoint,
                methods,
                rule,
            ))
            output.append(line)

        for line in sorted(output):
            print(line)



import flask

from pg_discuss import ext
from pg_discuss import queries
from pg_discuss import tables
from pg_discuss.db import db

import sqlalchemy as sa


class Voting(ext.AppExtBase, ext.OnPreCommentSerialize):
    """Extension to enable upvotes/downvotes on comments. Votes counts
    are stored directly on comment object for performance reasons, but
    :class:`pg_discuss.models.IdentityComment` are used to tie votes
    to an Identity.
    """

    def init_app(self, app):
        app.route('/comments/<int:comment_id>/upvote', methods=['POST'])(
            self.upvote)
        app.route('/comments/<int:comment_id>/downvote', methods=['POST'])(
            self.downvote)

    def upvote(self, comment_id):
        return self.vote(comment_id, vote_type='upvote')

    def downvote(self, comment_id):
        return self.vote(comment_id, vote_type='downvote')

    def vote(self, comment_id, vote_type):
        """Add a new vote and increase the upvotes counter.

        Note: Use`jsonb_set` function in 9.5 to do an in-place increment:

            update comment
            set custom_json = jsonb_set(
                custom_json,
                '{upvotes}',
                coalesce(
                ((custom_json->>'upvotes')::integer + 1)::text::jsonb,
                '1')
            )
            where id = %id;
        """

        identity_comment = {
            'identity_id': flask.g.identity['id'],
            'comment_id': comment_id,
            'rel_type': vote_type,
        }
        t = tables.identity_comment
        ins = (
            t.insert()
            .values(**identity_comment)
            .returning(*list(t.c))
        )
        keyname = "{0}s".format(vote_type)  # add 's' to pluralize vote_type
        t = tables.comment
        # Increment the vote type if already set, or set the initial value
        # to 1, using a COALESCE expression.
        coalesce = (
            "coalesce(((custom_json->>'{0}')::integer + 1)::text::jsonb,'1')"
            .format(keyname)
        )
        incremented = sa.func.jsonb_set(
            sa.text('custom_json'),
            sa.text("'{{{0}}}'".format(keyname)),
            sa.text(coalesce)
        )
        upd = (
            t.update()
            .where(t.c.id == comment_id)
            .values(custom_json=incremented)
            .returning(
                t.c.custom_json['upvotes'],
                t.c.custom_json['downvotes']
            )
        )
        statements = [ins, upd]
        stmt, bindparams = queries.cte_chain(statements)
        try:
            results = db.engine.execute(stmt, **bindparams).first()
        except sa.exc.IntegrityError:
            flask.abort(
                400,
                'Cannot {0} on comment: identity has already submitted {0}'
                .format(vote_type)
            )

        resp_obj = {
            'upvotes': results[0],
            'downvotes': results[1]
        }
        return flask.jsonify(resp_obj)

    def on_pre_comment_serialize(self, raw_comment, client_comment, **extras):
        client_comment['upvotes'] = (
            raw_comment['custom_json'].get('upvotes', 0)
        )
        client_comment['downvotes'] = (
            raw_comment['custom_json'].get('downvotes', 0)
        )



"""Module containing default and "blessed" JSON encoders/decoders.
"""
import datetime

import pytz
import simplejson as json

EPOCH = datetime.datetime(1970, 1, 1, 0, 0, 0, 0, tzinfo=pytz.utc)


class UnixTimeJSONEncoder(json.JSONEncoder):
    """JSON encoder driver to encode datetime values as Unix timestamps.
    """
    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            return (obj - EPOCH).total_seconds()
        return json.JSONEncoder.default(self, obj)



import voluptuous

from pg_discuss import ext

#: Minimum allowed comment length
MIN_COMMENT_LENGTH = 3
#: Maximum allowed comment length
MAX_COMMENT_LENGTH = 65535


class ValidateCommentLen(ext.ValidateComment):
    """Extension to add a validation rule that enforces a minimum and
    maximum comment length.
    """

    def __init__(self, app):
        app.config.setdefault('MIN_COMMENT_LENGTH', MIN_COMMENT_LENGTH)
        app.config.setdefault('MAX_COMMENT_LENGTH', MAX_COMMENT_LENGTH)
        super(ext.ValidateComment, self).__init__(app)

    def validate_comment(self, comment, action, **extras):
        text = comment['text']
        min_comment_length = self.app.config['MIN_COMMENT_LENGTH']
        max_comment_length = self.app.config['MAX_COMMENT_LENGTH']
        voluptuous.Length(min=min_comment_length, max=max_comment_length)(
            text.rstrip())
        return comment



"""Python 2/3 compatibility helpers.

Some parts from flask_wtf._compat:
https://github.com/lepture/flask-wtf/blob/HEAD/flask_wtf/csrf.py
"""
import sys
import urllib

try:
    from urlparse import urlparse
except ImportError:  # pragma: no cover
    # python 3
    from urllib.parse import urlparse  # NOQA

from functools import reduce

PY3 = sys.version_info[0] == 3
PYPY = hasattr(sys, 'pypy_version_info')

if PY3:  # pragma: no cover
    text_type = str
    string_types = (str,)
    unquote = urllib.parse.unquote
else:  # pragma: no cover
    text_type = unicode  # NOQA
    string_types = (str, unicode)  # NOQA
    unquote = urllib.unquote


def to_bytes(text):
    """Transform string to bytes."""
    if isinstance(text, text_type):
        text = text.encode('utf-8')
    return text


def to_unicode(input_bytes, encoding='utf-8'):
    """Decodes input_bytes to text if needed."""
    if not isinstance(input_bytes, string_types):
        input_bytes = input_bytes.decode(encoding)
    return input_bytes



import flask

from pg_discuss import ext


class CaptureRemoteAddr(ext.ValidateComment):
    """Extension to capture and persist the IP address associated with comment
    submissions.
    """
    def validate_comment(self, comment, action, **extras):
        comment['custom_json']['remote_addr'] = flask.request.remote_addr
        return comment



"""empty message

Revision ID: 46a03f51647
Revises: None
Create Date: 2015-09-27 20:56:54.697804

"""

# revision identifiers, used by Alembic.
revision = '46a03f51647'
down_revision = None

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

def upgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.create_table('admin_user',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('login', sa.String(), nullable=False),
    sa.Column('email', sa.String(), nullable=False),
    sa.Column('active', sa.Boolean(), nullable=False),
    sa.Column('password', sa.String(), nullable=False),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('login')
    )
    op.create_table('identity',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('created', sa.DateTime(timezone=True), server_default=sa.text('NOW()'), nullable=False),
    sa.Column('custom_json', postgresql.JSONB(), server_default='{}', nullable=False),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('thread',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('client_id', sa.String(), nullable=True),
    sa.Column('created', sa.DateTime(timezone=True), server_default=sa.text('NOW()'), nullable=False),
    sa.Column('custom_json', postgresql.JSONB(), server_default='{}', nullable=False),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('client_id')
    )
    op.create_table('comment',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('identity_id', sa.Integer(), nullable=True),
    sa.Column('thread_id', sa.Integer(), nullable=False),
    sa.Column('parent_id', sa.Integer(), nullable=True),
    sa.Column('version_of_id', sa.Integer(), nullable=True),
    sa.Column('created', sa.DateTime(timezone=True), server_default=sa.text('NOW()'), nullable=False),
    sa.Column('modified', sa.DateTime(timezone=True), server_default=sa.text('NOW()'), nullable=False),
    sa.Column('text', sa.String(), nullable=False),
    sa.Column('custom_json', postgresql.JSONB(), server_default='{}', nullable=False),
    sa.ForeignKeyConstraint(['identity_id'], ['identity.id'], ),
    sa.ForeignKeyConstraint(['parent_id'], ['comment.id'], ),
    sa.ForeignKeyConstraint(['thread_id'], ['thread.id'], ),
    sa.ForeignKeyConstraint(['version_of_id'], ['comment.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('identity_comment',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('identity_id', sa.Integer(), nullable=False),
    sa.Column('comment_id', sa.Integer(), nullable=False),
    sa.Column('rel_type', sa.String(), nullable=False),
    sa.Column('created', sa.DateTime(timezone=True), server_default=sa.text('NOW()'), nullable=False),
    sa.Column('custom_json', postgresql.JSONB(), server_default='{}', nullable=False),
    sa.ForeignKeyConstraint(['comment_id'], ['comment.id'], ),
    sa.ForeignKeyConstraint(['identity_id'], ['identity.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    ### end Alembic commands ###


def downgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('identity_comment')
    op.drop_table('comment')
    op.drop_table('thread')
    op.drop_table('identity')
    op.drop_table('admin_user')
    ### end Alembic commands ###
